import abcEconomics
from abcEconomics.agents import Household


class Killer(abcEconomics.Agent, Household):
    def init(self):
        # your agent initialization goes here, not in __init__
        pass

    def kill_silent(self):
        return ('victim', self.time)

    def kill_loud(self):
        return ('loudvictim', self.time)



from __future__ import division
import abcEconomics


class MessageA(abcEconomics.Agent):
    def init(self):
        # your agent initialization goes here, not in __init__
        pass

    def sendmsg(self):
        self.send(('messageb', self.id), 'msg', 'hello there')

    def recvmsg(self):
        msg = self.get_messages('msg')[0]
        assert msg == 'hello there'



import traceback
import os
from .plot_with_plotly import generate_graph_layout
import dash
import dash_core_components as dcc
import dash_html_components as html
import numpy as np
from dash.dependencies import Input, Output, State


def gen(simulation, parameter_mask, names, title, header, truncate_rounds, texts, pages):
    form = [html.P("scroll down to start")]

    inputs = []
    converter = []

    for parameter, value in list(parameter_mask.items()):
        try:
            title = names[parameter]
        except KeyError:
            title = parameter

        if value is None:
            form.append(dcc.Markdown(title))
            continue

        inputs.append(State(parameter, 'value'))

        if isinstance(value, bool):
            form.append(html.Div(children=[title, dcc.RadioItems(id=parameter,
                                                                 options=[{'label': 'True', 'value': True},
                                                                          {'label': 'False', 'value': False}],
                                                                 value=value,
                                                                 labelStyle={'display': 'inline-block'})]))
            converter.append(bool)

        elif isinstance(value, list):
            options = []
            for option in value:
                options.append({'label': option, 'value': option})

            form.append(html.Div(children=[title, dcc.RadioItems(id=parameter,
                                                                 options=options,
                                                                 value=value[0])]))
            converter.append(lambda x: x)

        elif isinstance(value, str):
            form.append(html.Div(children=[title, dcc.Textarea(id=parameter,
                                     placeholder=title,
                                     value=value)]))
            converter.append(str)

        elif isinstance(value, int):
            form.append(html.Div(children=[title,
                                           dcc.Input(id=parameter,
                                                     placeholder=value,
                                                     type='numeric',
                                                     value=value)]))
            converter.append(int)

        elif isinstance(value, float):
            form.append(html.Div(children=[title,
                                           dcc.Input(id=parameter,
                                                     placeholder=value,
                                                     type='numeric',
                                                     value=value)]))
            converter.append(float)
        else:

            if isinstance(value, tuple):
                try:
                    min_value, default, max_value, step = value
                except ValueError:
                    min_value, default, max_value = value
                    step = (max_value - min_value) / 100

                if (isinstance(min_value, float) or
                    isinstance(default, float) or
                    isinstance(max_value, float) or
                    isinstance(step, float)):
                        dspstep = (max_value - min_value) / 10
                        form.append(html.Div(children=[title,
                            dcc.Slider(id=parameter,
                                       min=min_value, max=max_value, step=step, value=default,
                                       marks={n: str(n)
                                              for n in np.arange(min_value, max_value, dspstep)})]))
                        converter.append(float)
                else:
                    dspstep = max(1, int((max_value - min_value) / 10))
                    form.append(html.Div(children=[title,
                                                   dcc.Slider(id=parameter,
                                                              min=min_value, max=max_value, step=step, value=default,
                                                              marks={n: str(n)
                                                                     for n in range(min_value, max_value, dspstep)})]))
                    converter.append(int)


    form.append(html.Button('Run Simulation', id='run_btn', type="submit"))

    form.append(html.Div(children=[dcc.Textarea(id='saveas',
                                                placeholder="Save as",
                                                value='abcEconomicsSim'),
                                   html.Button('Save', id='save_btn')]))

    try:
        graphs = generate_graph_layout('')
    except FileNotFoundError:
        graphs = html.Div()

    layout = html.Div(children=[
        html.H1(children=title),
        dcc.Markdown(header),
        dcc.Tabs(id='graphs', children=[
                 dcc.Tab(label='Run', children=html.Div(children=form), value=10),
                 dcc.Tab(id='result', label='Result', children=graphs, value=1)],
                 value=10)])

    return layout, inputs, converter


def newest_subdirectory(directory='.', name=''):
    """ Returns the newes subdirectory in the 'directory/name' directory """
    directory = os.path.abspath(directory)
    all_subdirs = [os.path.join(directory, name)
                   for name in os.listdir(directory)
                   if os.path.isdir(os.path.join(directory, name))]
    all_subdirs = sorted(all_subdirs, key=os.path.getmtime, reverse=True)
    for subdir in all_subdirs:
        if name in subdir:
            return subdir + '/'
    raise Exception('No result/* directory')


def dashgui(simulation, parameter_mask, names, title, header, truncate_rounds, texts, pages):
    app = dash.Dash()

    app.layout, inputs, converter = gen(simulation, parameter_mask, names, title, header, truncate_rounds, texts, pages)

    @app.callback(Output('result', 'children'),
                  [Input('run_btn', 'n_clicks')],
                  inputs)
    def call_simulation(clicks, *parameters):
        print("HERE")
        if clicks is not None:
            params = []
            for i, param in enumerate(parameters):
                try:
                    params.append(converter[i](param))
                except Exception as e:
                    problem = "Parameter %s not of type %s: %s " % (list(parameter_mask.keys())[i], str(converter[i]), str(e))
                    print(problem)
                    return html.Div(children=[problem])

            params = dict(zip(parameter_mask.keys(), params))
            print(params)
            try:
                simulation(params)
            except Exception as e:
                traceback.print_exc()
                return html.Div(children=["The simulation produced an error with this particular parameterisation", str(type(e)), str(e)])

            return generate_graph_layout(newest_subdirectory('./result', ''))
        else:
            return html.Div()

    app.run_server(debug=True, use_reloader=False)



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# abcEconomics is open-source software. If you are using abcEconomics for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
The :class:`abcEconomicsagent.Agent` class is the basic class for creating your agent. It
automatically handles the possession of goods of an agent. In order to produce/transform
goods you need to also subclass the :class:`abcEconomicsagent.Firm` [1]_ or to create a consumer
the :class:`abcEconomicsagent.Household`.

For detailed documentation on:

Trading:
    see :class:`abcEconomicsagent.Trade`
Logging and data creation:
    see :class:`abcEconomicsagent.Database` and :doc:`simulation_results`
Messaging between agents:
    see :class:`abcEconomicsagent.Messenger`.

.. autoexception:: abcEconomicstools.NotEnoughGoods

.. [1] or :class:`abcEconomicsagent.FirmMultiTechnologies` for simulations with complex technologies.
"""
# ***************************************************************************************** #
#  trade.pyx is written in cython. When you modify trade.pyx you need to compile it with    #
# compile.sh and compile.py because the resulting trade.c file is distributed.              #
# Don't forget to commit it to git                                                          #
# ***************************************************************************************** #
import random
from collections import defaultdict, OrderedDict
from abcEconomics.notenoughgoods import NotEnoughGoods

epsilon = 0.00000000001


def get_epsilon():
    return epsilon


class Offer(object):
    __slots__ = ('sender', 'receiver', 'good', 'quantity', 'price', 'currency',
                 'sell', 'status', 'final_quantity', 'id',
                 'made', 'status_round')
    """ This is an offer container that is send to the other agent. You can
    access the offer container both at the receiver as well as at the sender,
    if you have saved the offer. (e.G. self.offer = self.sell(...))

    it has the following properties:
        sender:
            this is the name of the sender

        receiver:
            this is the name of the receiver

        currency:
            The other good against which the good is traded.

        good:
            the good offered or demanded

        quantity:
            the quantity offered or demanded

        price:
            the suggested transaction price

        sell:
            this can have the values False for buy; True for sell

        status:
            'new':
                has been created, but not answered

            'accepted':
                trade fully accepted

            'rejected':
                trade rejected

            'pending':
                offer has not yet answered, and is not older than one round.

            'perished':
                the **perishable** good was not accepted by the end of the round
                and therefore perished.

        final_quantity:
            If the offer has been answered this returns the actual quantity
            bought or sold. (Equal to quantity if the offer was accepted fully)
        id:
            a unique identifier
    """
    def __init__(self, sender, receiver, good, quantity, price, currency,
                 sell, id, made):
        self.sender = sender
        self.receiver = receiver
        self.good = good
        self.currency = currency
        self.quantity = quantity
        self.price = price
        self.sell = sell
        self.status = 'new'
        self.final_quantity = None
        self.id = id
        self.made = made
        self.status_round = None

    def __repr__(self):
        return ("""<{sender: %s, receiver: %s, good: %s, quantity: %f, price: %f, currency: %f,
                sell: %s, status: %s, final_quantity: % f, id: %i, made: %s, status_round: %s }>""" %
                (self.sender, self.receiver, self.good, self.quantity, self.price, self.currency,
                 self.sell, self.status, self.final_quantity, self.id, self.made, self.status_round))


class Trade:
    """ Agents can trade with each other. The clearing of the trade is taken care
    of fully by abcEconomics.
    Selling a good works in the following way:

    1. An agent sends an offer. :meth:`~.sell`

       *abcEconomics does not allow you to sell the same good twice; self.free(good) shows how much good is not reserved yet*

    2. **Next subround:** An agent receives the offer :meth:`~.get_offers`, and can
       :meth:`~.accept`, :meth:`~.reject` or partially accept it. :meth:`~.accept`

       *The good is credited and the price is deducted from the agent's possessions.*

    3. **Next subround:**

       - in case of acceptance *the money is automatically credited.*
       - in case of partial acceptance *the money is credited and part of the reserved good is unblocked.*
       - in case of rejection *the good is unblocked.*

    Analogously for buying: :meth:`~.buy`

    Example::

        # Agent 1
        def sales(self):
            self.remember_trade = self.sell('Household', 0, 'cookies', quantity=5, price=self.price)

        # Agent 2
        def receive_sale(self):
            oo = self.get_offers('cookies')
            for offer in oo:
                if offer.price < 0.3:
                    try:
                        self.accept(offer)
                    except NotEnoughGoods:
                        self.accept(offer, self['money'] / offer.price)
                else:
                    self.reject(offer)

        # Agent 1, subround 3
        def learning(self):
            offer = self.info(self.remember_trade)
            if offer.status == 'reject':
                self.price *= .9
            elif offer.status = 'accepted':
                self.price *= offer.final_quantity / offer.quantity
    Example::

        # Agent 1
        def sales(self):
            self.remember_trade = self.sell('Household', 0, 'cookies', quantity=5, price=self.price, currency='dollars')

        # Agent 2
        def receive_sale(self):
            oo = self.get_offers('cookies')
            for offer in oo:
                if ((offer.currency == 'dollars' and offer.price < 0.3 * exchange_rate)
                    or (offer.currency == 'euros' and dollars'offer.price < 0.3)):

                    try:
                        self.accept(offer)
                    except NotEnoughGoods:
                        self.accept(offer, self['money'] / offer.price)
                else:
                    self.reject(offer)

    If we did not implement a barter class, but one can use this class as a barter class,
    """
    def __init__(self, id, agent_parameters, simulation_parameters):
        super(Trade, self).__init__(id, agent_parameters, simulation_parameters)
        # unpack simulation_parameters
        trade_logging = simulation_parameters['trade_logging']

        self.given_offers = OrderedDict()
        self._open_offers_buy = defaultdict(dict)
        self._open_offers_sell = defaultdict(dict)
        self._polled_offers = {}
        self._offer_count = 0
        self.trade_logging = {'individual': 1,
                              'group': 2, 'off': 0}[trade_logging]
        self._trade_log = defaultdict(int)
        self._quotes = {}

    def _offer_counter(self):
        """ returns a unique number for an offer (containing the agent's name)
        """
        self._offer_count += 1
        return hash((self.name, self._offer_count))

    def _advance_round(self, time):
        if self.trade_logging > 0:
            self.database_connection.put(["trade_log", self._trade_log, self.time])
        self._trade_log = defaultdict(int)

    def get_buy_offers_all(self, descending=False, sorted=True):
        """ """
        goods = list(self._open_offers_buy.keys())
        return {good: self.get_buy_offers(good, descending, sorted) for good in goods}

    def get_sell_offers_all(self, descending=False, sorted=True):
        """ """
        goods = list(self._open_offers_sell.keys())
        return {good: self.get_sell_offers(good, descending, sorted) for good in goods}

    def get_offers_all(self, descending=False, sorted=True):
        """ returns all offers in a dictionary, with goods as key. The in each
        goods-category the goods are ordered by price. The order can be reversed
        by setting descending=True

        *Offers that are not accepted in the same subround (def block) are
        automatically rejected.* However you can also manually reject.

        Args:

         descending(optional):
            is a bool. False for descending True for ascending by price

         sorted(default=True):
                Whether offers are sorted by price. Faster if False.

        Returns:

            a dictionary with good types as keys and list of :class:`abcEconomics.trade.Offer`
            as values

        Example::

         oo = get_offers_all(descending=False)
         for good_category in oo:
            print('The cheapest good of category' + good_category
            + ' is ' + good_category[0])
            for offer in oo[good_category]:
                if offer.price < 0.5:
                    self.accept(offer)

         for offer in oo.beer:
            print(offer.price, offer.sender)
        """
        goods = list(self._open_offers_sell.keys() + self._open_offers_buy.keys())
        return {good: self.get_offers(good, descending, sorted) for good in goods}

    def get_buy_offers(self, good, sorted=True, descending=False, shuffled=True):
        """ """
        ret = list(self._open_offers_buy[good].values())
        self._polled_offers.update(self._open_offers_buy[good])
        del self._open_offers_buy[good]
        if shuffled:
            random.shuffle(ret)
        if sorted:
            ret.sort(key=lambda objects: objects.price, reverse=descending)
        return ret

    def get_sell_offers(self, good, sorted=True, descending=False, shuffled=True):
        """ """
        ret = list(self._open_offers_sell[good].values())
        self._polled_offers.update(self._open_offers_sell[good])
        del self._open_offers_sell[good]
        if shuffled:
            random.shuffle(ret)
        if sorted:
            ret.sort(key=lambda objects: objects.price, reverse=descending)
        return ret

    def get_offers(self, good, sorted=True, descending=False, shuffled=True):
        """ returns all offers of the 'good' ordered by price.

        *Offers that are not accepted in the same subround (def block) are
        automatically rejected.* However you can also manually reject.

        peek_offers can be used to look at the offers without them being
        rejected automatically

        Args:
            good:
                the good which should be retrieved

            sorted(bool, default=True):
                Whether offers are sorted by price. Faster if False.

            descending(bool, default=False):
                False for descending True for ascending by price

            shuffled(bool, default=True):
                whether the order of messages is randomized or correlated with
                the ID of the agent. Setting this to False speeds up the
                simulation considerably, but introduces a bias.

        Returns:
            A list of :class:`abcEconomics.trade.Offer` ordered by price.

        Example::

            offers = get_offers('books')
            for offer in offers:
                if offer.price < 50:
                    self.accept(offer)
                elif offer.price < 100:
                    self.accept(offer, 1)
                else:
                    self.reject(offer)  # optional
        """
        ret = (self.get_buy_offers(good, descending=False, sorted=False, shuffled=False) +
               self.get_sell_offers(good, descending=False, sorted=False, shuffled=False))
        if shuffled:
            random.shuffle(ret)
        if sorted:
            ret.sort(key=lambda objects: objects.price, reverse=descending)
        return ret

    def peak_buy_offers(self, good, sorted=True, descending=False, shuffled=True):
        """ """
        ret = []
        for offer in self._open_offers_buy[good].values():
            ret.append(offer)
        if shuffled:
            random.shuffle(ret)
        if sorted:
            ret.sort(key=lambda objects: objects.price, reverse=descending)
        return ret

    def peak_sell_offers(self, good, sorted=True, descending=False, shuffled=True):
        """ """
        ret = []
        for offer in self._open_offers_sell[good].values():
            ret.append(offer)
        if shuffled:
            random.shuffle(ret)
        if sorted:
            ret.sort(key=lambda objects: objects.price, reverse=descending)
        return ret

    def peak_offers(self, good, sorted=True, descending=False, shuffled=True):
        """ returns a peak on all offers of the 'good' ordered by price.
        Peaked offers can not be accepted or rejected and they do not
        expire.

        Args:
            good:
                the good which should be retrieved
                descending(bool, default=False):
                False for descending True for ascending by price

        Returns:
            A list of offers ordered by price

        Example::

            offers = get_offers('books')
            for offer in offers:
                if offer.price < 50:
                    self.accept(offer)
                elif offer.price < 100:
                    self.accept(offer, 1)
                else:
                    self.reject(offer)  # optional
        """
        ret = (self.peak_buy_offers(good, sorted=False, descending=False, shuffled=False) +
               self.peak_sell_offers(good, sorted=False, descending=False, shuffled=False))
        if shuffled:
            random.shuffle(ret)
        if sorted:
            ret.sort(key=lambda objects: objects.price, reverse=descending)
        return ret

    def sell(self, receiver,
             good, quantity, price, currency='money', epsilon=epsilon):
        """ Sends a offer to sell a particular good to somebody. The amount promised
        is reserved. (self.free(good), shows the not yet reserved goods)

        Args:
            receiver:
                the receiving agent

            'good':
                name of the good

            quantity:
                maximum units disposed to buy at this price

            price:
                price per unit

            currency:
                is the currency of this transaction (defaults to 'money')

            epsilon (optional):
                if you have floating point errors, a quantity or prices is
                a fraction of number to high or low. You can increase the
                floating point tolerance. See troubleshooting -- floating point problems

        Returns:
            A reference to the offer. The offer and the offer status can
            be accessed with `self.info(offer_reference)`.

        Example::

            def subround_1(self):
                self.offer = self.sell('household', 1, 'cookies', quantity=5, price=0.1)

            def subround_2(self):
                offer = self.info(self.offer)
                if offer.status == 'accepted':
                    print(offer.final_quantity , 'cookies have be bougth')
                else:
                    offer.status == 'rejected':
                    print('On diet')
        """
        assert price > - epsilon, 'price %.30f is smaller than 0 - epsilon (%.30f)' % (price, - epsilon)
        if price < 0:
            price = 0
        # makes sure the quantity is between zero and maximum available, but
        # if its only a little bit above or below its set to the bounds
        assert quantity > - epsilon, 'quantity %.30f is smaller than 0 - epsilon (%.30f)' % (quantity, - epsilon)
        if quantity < 0:
            quantity = 0

        self._inventory.reserve(good, quantity)

        offer_id = self._offer_counter()
        offer = Offer(self.name,
                      receiver,
                      good,
                      quantity,
                      price,
                      currency,
                      True,
                      offer_id,
                      self.time)
        self.given_offers[offer_id] = offer
        self.send(receiver, 'abcEconomics_propose_sell', offer)
        return offer

    def buy(self, receiver, good,
            quantity, price, currency='money', epsilon=epsilon):
        """ Sends a offer to buy a particular good to somebody. The money promised
        is reserved. (self.free(currency), shows the not yet reserved goods)

        Args:
            receiver:
                The name of the receiving agent a tuple (group, id).
                e.G. ('firm', 15)

            'good':
                name of the good

            quantity:
                maximum units disposed to buy at this price

            price:
                price per unit

            currency:
                is the currency of this transaction (defaults to 'money')

            epsilon (optional):
                if you have floating point errors, a quantity or prices is
                a fraction of number to high or low. You can increase the
                floating point tolerance. See troubleshooting -- floating point problems
        """
        assert price > - epsilon, 'price %.30f is smaller than 0 - epsilon (%.30f)' % (price, - epsilon)
        if price < 0:
            price = 0
        # makes sure the money_amount is between zero and maximum available, but
        # if its only a little bit above or below its set to the bounds

        assert quantity > - epsilon, (
            '%s quantity %.30f is smaller than 0 - epsilon (%.30f)' % (currency, quantity, - epsilon))
        if quantity < 0:
            quantity = 0

        offer_id = self._offer_counter()
        self._inventory.reserve(currency, quantity * price)
        offer = Offer(self.name,
                      receiver,
                      good,
                      quantity,
                      price,
                      currency,
                      False,
                      offer_id,
                      self.time)
        self.send(receiver, 'abcEconomics_propose_buy', offer)
        self.given_offers[offer_id] = offer
        return offer

    def accept(self, offer, quantity=-999, epsilon=epsilon):
        """ The buy or sell offer is accepted and cleared. If no quantity is
        given the offer is fully accepted; If a quantity is given the offer is
        partial accepted.

        Args:

            offer:
                the offer the other party made
            quantity:
                quantity to accept. If not given all is accepted

            epsilon (optional):
                if you have floating point errors, a quantity or prices is
                a fraction of number to high or low. You can increase the
                floating point tolerance. See troubleshooting -- floating point problems

        Return:
            Returns a dictionary with the good's quantity and the amount paid.
        """
        offer_quantity = offer.quantity
        if quantity == -999:
            quantity = offer_quantity
        assert quantity > - epsilon, 'quantity %.30f is smaller than 0 - epsilon (%.30f)' % (quantity, - epsilon)
        if quantity < 0:
            quantity = 0
        if quantity > offer_quantity + epsilon * max(quantity, offer_quantity):
            raise AssertionError('accepted more than offered %s: %.100f >= %.100f'
                                 % (offer.good, quantity, offer_quantity))
        if quantity > offer_quantity:
            quantity = offer_quantity

        if quantity == 0:
            self.reject(offer)
            return {offer.good: 0, offer.currency: 0}

        money_amount = quantity * offer.price
        if offer.sell:  # ord('s')
            assert money_amount > - epsilon, 'money = quantity * offer.price %.30f is smaller than 0 - epsilon (%.30f)' % (money_amount, - epsilon)
            if money_amount < 0:
                money_amount = 0

            available = self._inventory[offer.currency]
            if money_amount > available + epsilon + epsilon * max(money_amount, available):
                raise NotEnoughGoods(self.name, offer.currency, money_amount - available)
            if money_amount > available:
                money_amount = available
            self._inventory.haves[offer.good] += quantity
            self._inventory.haves[offer.currency] -= quantity * offer.price
        else:
            assert quantity > - epsilon, 'quantity %.30f is smaller than 0 - epsilon (%.30f)' % (quantity, - epsilon)
            if quantity < 0:
                quantity = 0
            available = self._inventory[offer.good]
            if quantity > available + epsilon + epsilon * max(quantity, available):
                raise NotEnoughGoods(self.name, offer.good, quantity - available)
            if quantity > available:
                quantity = available
            self._inventory.haves[offer.good] -= quantity
            self._inventory.haves[offer.currency] += quantity * offer.price
        offer.final_quantity = quantity
        self.send(offer.sender, 'abcEconomics_receive_accept', (offer.id, quantity))
        del self._polled_offers[offer.id]
        if offer.sell:
            return {offer.good: - quantity, offer.currency: money_amount}
        else:
            return {offer.good: quantity, offer.currency: - money_amount}

    def _reject_polled_but_not_accepted_offers(self):
        for offer in self._polled_offers.values():
            self._reject(offer)
        self._polled_offers = {}

    def _reject(self, offer):
        """  Rejects the offer offer

        Args:
            offer:
                the offer the other party made
                (offer not quote!)
        """
        self.send(offer.sender, 'abcEconomics_receive_reject', offer.id)

    def reject(self, offer):
        """ Rejects and offer, if the offer is subsequently accepted in the
        same subround it is accepted'. Peaked offers can not be rejected.

        Args:

            offer:
                the offer to be rejected
        """
        pass

    def _receive_accept(self, offer_id_final_quantity):
        """ When the other party partially accepted the  money or good is
        received, remaining good or money is added back to haves and the offer
        is deleted
        """
        offer = self.given_offers[offer_id_final_quantity[0]]
        offer.final_quantity = offer_id_final_quantity[1]
        if offer.sell:
            self._inventory.commit(offer.good, offer.quantity, offer.final_quantity)
            self._inventory.haves[offer.currency] += offer.final_quantity * offer.price
        else:
            self._inventory.haves[offer.good] += offer.final_quantity
            self._inventory.commit(offer.currency, offer.quantity * offer.price, offer.final_quantity * offer.price)
        offer.status = "accepted"
        offer.status_round = self.time
        del self.given_offers[offer.id]
        return offer

    def _log_receive_accept_group(self, offer):
        if offer.sell:
            self._trade_log[(offer.good, self.group, offer.receiver[0], offer.price)] += offer.final_quantity
        else:
            self._trade_log[(offer.good, offer.receiver[0], self.group, offer.price)] += offer.final_quantity

    def _log_receive_accept_agent(self, offer):
        if offer.sell:
            self._trade_log[(offer.good, self.name_without_colon, '%s_%i' % (*offer.receiver, offer.price))] += offer.final_quantity
        else:
            self._trade_log[(offer.good, '%s_%i' % (*offer.receiver, self.name_without_colon, offer.price))] += offer.final_quantity

    def _receive_reject(self, offer_id):
        """ deletes a given offer

        is used by _do_message_clearing, when the other party rejects
        or at the end of the subround when agent retracted the offer

        """
        offer = self.given_offers[offer_id]
        if offer.sell:
            self._inventory.rewind(offer.good, offer.quantity)
        else:
            self._inventory.rewind(offer.currency, offer.quantity * offer.price)
        offer.status = "rejected"
        offer.status_round = self.time
        offer.final_quantity = 0
        del self.given_offers[offer_id]

    def _delete_given_offer(self, offer_id):
        offer = self.given_offers.pop(offer_id)
        if offer.sell:
            self._inventory.rewind(offer.good, offer.quantity)
        else:
            self._inventory.rewind(offer.currency, offer.quantity * offer.price)

    def give(self, receiver, good, quantity, epsilon=epsilon):
        """ gives a good to another agent

        Args:
            receiver:
                The name of the receiving agent a tuple (group, id).
                e.G. ('firm', 15)
            good:
                the good to be transfered
            quantity:
                amount to be transfered

            epsilon (optional):
                if you have floating point errors, a quantity or prices is
                a fraction of number to high or low. You can increase the
                floating point tolerance. See troubleshooting -- floating point problems

        Raises:

            AssertionError, when good smaller than 0.

        Return:
            Dictionary, with the transfer, which can be used by self.log(...).

        Example::

            self.log('taxes', self.give('money': 0.05 * self['money'])

        """
        assert quantity > - epsilon, 'quantity %.30f is smaller than 0 - epsilon (%.30f)' % (quantity, - epsilon)
        if quantity < 0:
            quantity = 0
        available = self._inventory[good]
        if quantity > available + epsilon + epsilon * max(quantity, available):
            raise NotEnoughGoods(self.name, good, quantity - available)
        if quantity > available:
            quantity = available
        self._inventory.haves[good] -= quantity
        self.send(receiver, 'abcEconomics_receive_good', [good, quantity])
        return {good: quantity}

    def take(self, receiver, good, quantity, epsilon=epsilon):
        """ take a good from another agent. The other agent has to accept.
        using self.accept()

        Args:


            receiver:
                the receiving agent

            good:
                the good to be taken

            quantity:
                the quantity to be taken

            epsilon (optional):
                if you have floating point errors, a quantity or prices is
                a fraction of number to high or low. You can increase the
                floating point tolerance. See troubleshooting -- floating point problems
        """
        self.buy(receiver, good=good, quantity=quantity, price=0, epsilon=epsilon)


def compare_with_ties(x, y):
    if x < y:
        return -1
    elif x > y:
        return 1
    else:
        return random.randint(0, 1) * 2 - 1



class OnlineVariance:
    def __init__(self):
        self.clear()

    def clear(self):
        self.n = 0
        self.M2 = 0.0
        self._mean = 0.0

    def update(self, x):
        self.n += 1
        delta = x - self._mean
        self._mean += delta / self.n
        delta2 = x - self._mean
        self.M2 += delta * delta2

    def std(self):
        if self.n < 2:
            return 0.0
        else:
            return (self.M2 / (self.n - 1)) ** 0.5

    def mean(self):
        return self._mean

    def sum(self):
        return self._mean * self.n



""" Describes abcEconomics on GUI, if no other text is given"""


abcEconomicsdescription = """ (drag me)
<h1>abcEconomics a Python based modeling library for economic simulations.</h1>
For simulations of trade, production and consumption, abcEconomics comes with stan
dard functions that implement these kinds of interactions and actions. The
modeler only has to implement the logic and decisions of an agent; abcEconomics takes
care of all exchange of goods and production and consumption.







One special feature of abcEconomics is that goods have the physical properties of


goods in reality. In other words if agent A gives a good to agent B, then -
 unlike information - agent B receives the good and agent B does not have the
 good anymore. That means that agents can trade, produce or consume a good.
 The ownership and transformations (production or consumption) of goods are
 automatically handled by the platform.







abcEconomics models are programmed in standard Python, stock functions of agents a


re inherited from archetype classes (Agent, Firm or Household). The only no
t-so-standard Python is that agents are executed in parallel by the
Simulation class (in start.py).







abcEconomics allows the modeler to program agents as ordinary Python class-objects


, but run the simulation on a multi-core/processor computer. It takes no ef
fort or intervention from the modeller to run the simulation on a multi-core
processor production, consumption, trade, communication and similar
functions are automatically handled by the platform. The modeler only needs
to instruct abcEconomics, which automatically executes the specific functions.







abcEconomics is a scheduler [1] and a set of agent classes. According to the sched


ule the simulation class calls - each sub-round - agents to execute some ac
tions. Each agent executes these actions using some of the build-in
functions, such as trade, production and consumption of abcEconomics. The agents
can use the full set of commands of the Python general purpose language.







The audience of abcEconomics are economists that want to model agent-based models


of trade and production. It is especially geared towards simulations that a
re similar to standard economic models like general or partial equilibrium
models [2]. What is more abcEconomics is especially designed to make writing the
simulation and the execution fast. Therefore models can be developed in an
interlinked process of running and rewriting the simulation.







abcEconomics uses Python - a language that is especially beginner friendly, but al

so easy to learn for people who already know object oriented programming la
nguages such as Java, C++ or even MATLAB. Python allows simple, but fully
functional, programming for economists. What is more Python is readable even
for non Python programmers.
<h2> <a href="http://abcEconomics.readthedocs.io/en/master/">Find the Documentatio
n at readthedocs</a> </h2>"""



from __future__ import division
import abcEconomics


class MessageB(abcEconomics.Agent):
    def init(self):
        # your agent initialization goes here, not in __init__
        pass

    def sendmsg(self):
        self.send(('messagea', self.id), 'msg', 'hello there')

    def recvmsg(self):
        assert self.get_messages('msg')[0] == 'hello there'



""" Copyright 2012 Davoud Taghawi-Nejad

 Module Author: Davoud Taghawi-Nejad

 abcEconomics is open-source software. If you are using abcEconomics for your research you are
 requested the quote the use of this software.

 Licensed under the Apache License, Version 2.0 (the "License"); you may not
 use this file except in compliance with the License and quotation of the
 author. You may obtain a copy of the License at
       http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 License for the specific language governing permissions and limitations under
 the License.
"""
# pylint: disable=W0212, C0111


import re
import multiprocessing as mp
from multiprocessing.managers import BaseManager
import traceback
from collections import defaultdict, ChainMap

from .singleprocess import SingleProcess


class MyManager(BaseManager):
    pass


class ProcessorGroup(SingleProcess):
    def __init__(self, batch, queues, processes):
        self.agents = {}
        self.batch = batch
        self.queues = queues
        self.queue = queues[self.batch]
        self.processes = processes

    def add_agents(self, Agent, simulation_parameters, agent_parameters, default_sim_params, maxid):
        """appends an agent to a group """
        if isinstance(agent_parameters, int):
            agent_parameters = ([] for _ in range(agent_parameters))

        names = {}
        for id, ap in enumerate(agent_parameters, maxid):
            agent = Agent(id, ap, {**default_sim_params, **simulation_parameters})
            agent.send = agent._send_multiprocessing
            agent._out = defaultdict(list)
            agent.init(**ChainMap(simulation_parameters, ap))
            if hash(agent.name) % self.processes == self.batch:
                assert agent.name not in self.agents, ('Two agents with the same name %s' % str(agent.name))
                agent._str_name = re.sub('[^0-9a-zA-Z_]', '', str(agent.name))
                names[agent.name] = agent.name
                agent._processes = self.processes
                self.agents[agent.name] = agent
        return names

    def do(self, names, command, args, kwargs):
        try:
            self.rets = []
            self.post = [[] for _ in range(self.processes)]
            for name in names:
                if hash(name) % self.processes == self.batch:
                    agent = self.agents[name]
                    ret = agent._execute(command, args, kwargs)
                    self.rets.append(ret)
                    pst = agent._post_messages_multiprocessing(self.processes)
                    for o in range(self.processes):
                        self.post[o].extend(pst[o])
        except Exception:
            traceback.print_exc()
            raise

    def post_messages(self, names):
        for i in range(self.processes):
            self.queues[i].put(self.post[i])

        for i in range(self.processes):
            for receiver, envelope in self.queue.get():
                try:
                    self.agents[receiver].inbox.append(envelope)
                except KeyError:
                    print(envelope)
                    raise KeyError("Receiver %s does not exist" % str(name))
        return self.rets


class MultiProcess(object):
    """ This is a container for all agents. It exists only to allow for multiprocessing with MultiProcess.
    """

    def __init__(self, processes):
        manager = mp.Manager()
        self.queues = [manager.Queue() for _ in range(processes)]
        self.pool = mp.Pool(processes)
        MyManager.register('ProcessorGroup', ProcessorGroup)
        self.processor_groups = []
        self.managers = []
        for i in range(processes):
            manager = MyManager()
            manager.start()
            self.managers.append(manager)
            pg = manager.ProcessorGroup(i, self.queues, processes)
            self.processor_groups.append(pg)

    def add_agents(self, Agent, simulation_parameters, agent_parameters, agent_arguments, maxid):
        """appends an agent to a group """
        names = self.pool.map(add_agents_wrapper, jkk(self.processor_groups,
                                                      Agent,
                                                      simulation_parameters,
                                                      agent_parameters,
                                                      agent_arguments,
                                                      maxid))

        return flatten(names)

    def delete_agents(self, names):
        self.pool.map(delete_agents_wrapper, jkk(self.processor_groups, names))

    def do(self, names, command, args, kwargs):
        self.pool.map(wrapper, jkk(self.processor_groups, names, command, args, kwargs))

    def post_messages(self, names):
        return flatten(self.pool.map(post_messages, jkk(self.processor_groups, names)))

    def advance_round(self, time, str_time):
        self.pool.map(advance_round_wrapper, jkk(self.processor_groups, time, str_time))

    def group_names(self):
        return self.processor_groups[0].group_names()


def wrapper(args):
    pg, names, command, args, kwargs = args
    pg.do(names, command, args, kwargs)


def post_messages(args):
    pg, names = args
    return pg.post_messages(names)


def add_agents_wrapper(arg):
    pg, Agent, simulation_parameters, agent_parameters, agent_arguments, maxid = arg
    return pg.add_agents(Agent, simulation_parameters, agent_parameters, agent_arguments, maxid)


def delete_agents_wrapper(arg):
    pg, names = arg
    pg.delete_agents(names)


def advance_round_wrapper(arg):
    pg, time, str_time = arg
    pg.advance_round(time, str_time)


def jkk(iterator, *args):
    for i in iterator:
        yield (i,) + args


def flatten(l):
    return [item for sublist in l for item in sublist]



import platform
import abcEconomics


class Returner(abcEconomics.Agent):
    def init(self):
        pass

    def returnit(self):
        return self.name, self.id, (0, 1, 2, 3), self.time


class Getter(abcEconomics.Agent):
    def init(self):
        pass

    def getit(self, zeug):
        assert set(zeug) == set([(('returner', 0), 0, (0, 1, 2, 3), self.time),
                                 (('returner', 1), 1, (0, 1, 2, 3), self.time),
                                 (('returner', 2), 2, (0, 1, 2, 3), self.time)]), list(zeug)

    def method(self, a, b, c, d, e, f):
        assert a == 'a'
        assert b == 'b'
        assert c == 'c'
        assert d == 'd'
        assert e == 'e'
        assert f == 'f'


def main(processes, rounds):
    sim = abcEconomics.Simulation(processes=processes)

    returners = sim.build_agents(Returner, 'returner', number=3)
    getters = sim.build_agents(Getter, 'getter', number=1)

    for y in range(rounds):
        sim.advance_round(y)
        ret = returners.returnit()
        getters.getit(ret)
        getters.method('a', 'b', 'c', d='d', f='f', e='e')
    sim.finalize()
    print('Returning tested \t\t\t\t\t\tOK')
    print('Calling with parameters tested \t\t\t\t\tOK')


if __name__ == '__main__':
    main(processes=1, rounds=30)
    if (platform.system() != 'Windows' and
            platform.python_implementation() != 'PyPy'):
        main(processes=4, rounds=30)



import random
from abcEconomics.agent import Agent
from abcEconomics import NotEnoughGoods
from tools import is_zero


class Buy(Agent):
    def init(self, rounds):
        self.last_round = rounds - 1
        self.tests = {'accepted': False, 'rejected': False, 'partial': False}
        self.price = 0
        if self.id == 1:
            self.tests['not_answered'] = False

    def one(self):
        """ Acts only if he is agent 0: sends an buy offer to agent 1 offer
        """
        if self.id % 2 == 0:
            self.create('money', random.uniform(0, 10000))
            self.money = self['money']
            self.price = random.uniform(0.0001, 1)
            quantity = random.uniform(0, self.money / self.price)
            self.offer = self.buy(('buy', self.id + 1),
                                  'cookies', quantity, self.price)
            assert self.not_reserved('money') == self.money - \
                quantity * self.price

    def two(self):
        """ Acts only if he is agent 1: recieves offers and accepts;
        rejects; partially accepts and leaves offers unanswerd.
        """
        if self.id % 2 == 1:
            self.create('cookies', random.uniform(0, 10000))
            cookies = self['cookies']
            oo = self.get_offers('cookies')
            assert oo
            for offer in oo:
                if random.randint(0, 10) == 0:
                    self.tests['not_answered'] = True
                    continue
                elif random.randint(0, 10) == 0:
                    self.reject(offer)
                    assert self['money'] == 0
                    assert self['cookies'] == cookies
                    self.tests['rejected'] = True
                    break  # tests the automatic clean-up of polled offers
                try:
                    self.accept(offer)
                    assert self['money'] == offer.price * offer.quantity
                    assert self['cookies'] == cookies - offer.quantity
                    self.tests['accepted'] = True
                except NotEnoughGoods:
                    self.accept(offer, self['cookies'])
                    assert is_zero(self['cookies'])
                    assert self['money'] == cookies * offer.price
                    self.tests['partial'] = True

    def three(self):
        """
        """
        if self.id % 2 == 0:
            offer = self.offer
            if offer.status == 'rejected':
                test = self.money - self['money']
                assert is_zero(test), test
                self.tests['rejected'] = True
            elif offer.status == 'accepted':
                if offer.final_quantity == offer.quantity:
                    assert self.money - offer.quantity * \
                        offer.price == self['money']

                    assert self['cookies'] == offer.quantity
                    self.tests['accepted'] = True
                else:
                    test = (self.money - offer.final_quantity *
                            offer.price) - self['money']
                    assert is_zero(test), test
                    test = self['cookies'] - offer.final_quantity
                    assert is_zero(test), test
                    self.tests['partial'] = True
            else:
                SystemExit('Error in buy')

    def laut(self):
        print("laut")
        self.haut = 'xxxx'

    def clean_up(self):
        self.destroy('money')
        self.destroy('cookies')

    def all_tests_completed(self):
        if self.time == self.last_round and self.id == 0:
            assert all(self.tests.values(
            )), 'not all tests have been run; abcEconomics workes correctly, restart the unittesting to do all tests %s' % self.tests
            print('Test abcEconomics.buy:\t\t\t\t\tOK')
            print('Test abcEconomics.accept\t(abcEconomics.buy):\t\tOK')
            print('Test abcEconomics.reject\t(abcEconomics.buy):\t\tOK')
            print('Test abcEconomics.accept, partial\t(abcEconomics.buy):\tOK')
            print('Test reject pending automatic \t(abcEconomics.buy):\tOK')



import abcEconomics
from abcEconomics.agents import Household

from tools import is_zero


class UtilityHousehold(abcEconomics.Agent, Household):
    def init(self, rounds):
        self.last_round = rounds - 1

        if self.id == 0 or self.id == 2:
            def utility_function(a, b, c):
                utility = max(a ** 0.2, b ** 0.5 * c ** 0.3)
                a = 0
                b = 0.9 * b
                return utility, locals()

            self.utility_function = utility_function

        elif self.id == 1 or self.id == 3:
            self.utility_function = self.create_cobb_douglas_utility_function(
                {'a': 0.2, 'b': 0.5, 'c': 0.3})

    def one(self):
        pass

    def two(self):
        pass

    def three(self):
        pass

    def clean_up(self):
        pass

    def consumption(self):
        if self.id == 0:
            self.create('a', 10)
            self.create('b', 10)
            self.create('c', 10)
            utility = self.consume(self.utility_function, {'a': 5, 'b': 3, 'c': 1})
            assert utility == max(5 ** 0.2, 3 ** 0.5 * 1 ** 0.3), utility
            assert self['a'] == 5
            assert self['b'] == 9.7
            assert self['c'] == 10
            self.destroy('a')
            self.destroy('b')
            self.destroy('c')

        elif self.id == 1:
            self.create('a', 10)
            self.create('b', 10)
            self.create('c', 10)
            utility = self.consume(self.utility_function, {'a': 5, 'b': 3, 'c': 1})
            assert utility == 5 ** 0.2 * 3 ** 0.5 * 1 ** 0.3, utility
            assert self['a'] == 5
            assert self['b'] == 7
            assert self['c'] == 9
            self.consume(self.utility_function, ['a', 'b', 'c'])
            assert self['a'] == 0
            assert self['b'] == 0
            assert self['c'] == 0

            pu = self.utility_function(**{'a': 5, 'b': 300, 'c': 10})
            assert pu == 5 ** 0.2 * 300 ** 0.5 * 10 ** 0.3

        elif self.id == 2:
            self.create('a', 10)
            self.create('b', 10)
            self.create('c', 10)
            utility = self.consume(self.utility_function, ['a', 'b', 'c'])
            assert is_zero(utility - max(10 ** 0.2, 10 ** 0.5 * 10 ** 0.3)
                           ), (utility, max(10 ** 0.2, 10 ** 0.5 * 10 ** 0.3))
            assert self['a'] == 0
            assert self['b'] == 9
            assert self['c'] == 10
            self.destroy('a')
            self.destroy('b')
            self.destroy('c')

        elif self.id == 3:
            self.create('a', 10)
            self.create('b', 10)
            self.create('c', 10)
            utility = self.consume(self.utility_function, ['a', 'b', 'c'])
            assert is_zero(utility - 10 ** 0.2 * 10 ** 0.5 * 10 **
                           0.3), (utility, 10 ** 0.2 * 10 ** 0.5 * 10 ** 0.3)
            assert self['a'] == 0
            assert self['b'] == 0
            assert self['c'] == 0

            pu = self.utility_function(**{'a': 5, 'b': 300, 'c': 10})
            assert pu == 5 ** 0.2 * 300 ** 0.5 * 10 ** 0.3

    def all_tests_completed(self):
        if self.time == self.last_round and self.id == 0:
            print('Test consume:                             \tOK')
            print('Test set_utility_function:                \tOK')
            print('Test set_cobb_douglas_utility_function    \tOK')



import abcEconomics


class MessageA(abcEconomics.Agent):
    def init(self):
        # your agent initialization goes here, not in __init__
        pass

    def sendmsg(self):
        self.send(('messageb', self.id), 'msg', 'hello there')

    def recvmsg(self):
        msg = self.get_messages('msg')[0]
        assert msg == 'hello there'


class MessageB(abcEconomics.Agent):
    def init(self):
        # your agent initialization goes here, not in __init__
        pass

    def sendmsg(self):
        self.send(('messagea', self.id), 'msg', 'hello there')

    def recvmsg(self):
        assert self.get_messages('msg')[0] == 'hello there'


def main(processes, rounds):
    s = abcEconomics.Simulation(processes=processes, name='unittest')

    print('build MessageA')
    messagea = s.build_agents(MessageA, 'messagea', 20)
    print('build MessageB')
    messageb = s.build_agents(MessageB, 'messageb', 20)

    for r in range(rounds):
        s.time = r
        (messagea + messageb).sendmsg()
        (messageb + messagea).recvmsg()

    print("Send and receive test:\t\tOK")
    s.finalize()


if __name__ == '__main__':
    main(processes=1, rounds=5)
    print('Iteration with 1 core finished')
    main(processes=4, rounds=5)
    print('Iteration with multiple processes finished')



from copy import copy


class Contracts(set):
    def __init__(self, par=set()):
        super(Contracts, self).__init__(par)

    def add(self, entry):
        assert entry not in self
        super().add(entry)

    def remove(self, entry):
        super().remove(entry)

    def possessions(self):
        return copy(self)

    def possession(self, typ):
        ret = set()
        for contract in self:
            if isinstance(contract, typ):
                ret.add(contract)
        return ret

    def calculate_netvalue(self, parameters, value_functions):
        return sum(value_functions[entry.__class__](entry, parameters)
                   for entry in self)

    def calculate_assetvalue(self, parameters, value_functions):
        return sum(max(value_functions[entry.__class__](entry, parameters), 0)
                   for entry in self)

    def calculate_liablityvalue(self, parameters, value_functions):
        return sum(min(value_functions[entry.__class__](entry, parameters), 0)
                   for entry in self)

    def calculate_valued_assets(self, parameters, value_functions):
        ret = {str(entry): value_functions[entry.__class__](entry, parameters)
               for entry in self
               if value_functions[entry.__class__](entry, parameters) >= 0}
        return ret

    def calculate_valued_liablities(self, parameters, value_functions):
        ret = {str(entry): value_functions[entry.__class__](entry, parameters)
               for entry in self
               if value_functions[entry.__class__](entry, parameters) < 0}
        return ret

    # contracts
    def _advance_round(self, round):
        du = set(filter(lambda c: c.terminated, self))
        self.difference_update(du)



import abcEconomics
import platform


class MyAgent(abcEconomics.Agent):
    def init(self, agent_parameters):
        print('init', agent_parameters)
        self.mine = agent_parameters
        self.diff = False

    def call_me_maybe(self):
        print(self.name, self.mine)


class DifferentAgent(abcEconomics.Agent):
    def init(self, agent_parameters):
        print('init', agent_parameters)
        self.mine = agent_parameters
        self.diff = True

    def call_me_maybe(self):
        print('I am different', self.name, self.mine)


def main(processes, rounds):
    sim = abcEconomics.Simulation()

    myagent = sim.build_agents(MyAgent, 'myagent',
                               agent_parameters=[{'agent_parameters': 0},
                                                 {'agent_parameters': 1},
                                                 {'agent_parameters': 2},
                                                 {'agent_parameters': 3},
                                                 {'agent_parameters': 4},
                                                 {'agent_parameters': 5},
                                                 {'agent_parameters': 6}])

    differentagent = sim.build_agents(DifferentAgent, 'differentagent',
                                      agent_parameters=[{'agent_parameters': 0},
                                                        {'agent_parameters': 1},
                                                        {'agent_parameters': 2},
                                                        {'agent_parameters': 3},
                                                        {'agent_parameters': 4},
                                                        {'agent_parameters': 5},
                                                        {'agent_parameters': 6}])

    print('len', len(myagent))

    sim.advance_round(0)
    myagent.call_me_maybe()
    print('simple')

    for r in range(1, 14):
        sim.advance_round(r)
        myagent[r % 7].call_me_maybe()
        print('--')

    print('two individuals')
    for r in range(14, 28):
        sim.advance_round(r)
        (myagent[r % 7] + myagent[3]).call_me_maybe()
        print('--')

    print('one class')
    for r in range(28, 42):
        sim.advance_round(r)
        myagent[r % 7, 3].call_me_maybe()
        print('--')

    print('whole class plus individual')
    for r in range(42, 56):
        sim.advance_round(r)
        (myagent + myagent[r % 7]).call_me_maybe()
        print('--')

    print('two classes')
    for r in range(56, 70):
        sim.advance_round(r)
        (myagent + differentagent).call_me_maybe()
        print('--')

    sim.finalize()


if __name__ == '__main__':
    main(processes=1, rounds=30)
    if (platform.system() != 'Windows' and
            platform.python_implementation() != 'PyPy'):
        main(processes=4, rounds=30)



# from distutils.core import setup
# from distutils.extension import Extension
from Cython.Distutils import build_ext

try:
    from setuptools import setup
    from setuptools import Extension
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension

setup(
    name='trade',
    ext_modules=[
        Extension('trade', ['trade.pyx']),
        Extension('logger/online_variance', ['logger/online_variance.pyx'])],
    cmdclass={'build_ext': build_ext}
)



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# abcEconomics is open-source software. If you are using abcEconomics for your research you
# are requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from sys import getsizeof
import datetime
import json
import os
import multiprocessing
import time
from collections import defaultdict
from .online_variance import OnlineVariance
from .postprocess import to_csv
import queue
import pandas as pd


class Database:
    """Separate thread that receives data from in_sok and saves it into a
    database"""

    def __init__(self, directory, name, trade_log, plugin=None, pluginargs=[]):
        super().__init__()

        # setting up directory
        self.directory = directory
        if directory is not None:
            os.makedirs(os.path.abspath('.') + '/result/', exist_ok=True)
            if directory == 'auto':
                self.directory = (os.path.abspath('.') + '/result/' + name + '_' +
                             datetime.datetime.now().strftime("%Y-%m-%d_%H-%M"))
                """ the directory variable contains the directory of the simulation outcomes
                it can be used to generate your own graphs as all resulting
                csv files are there.
                """
            else:
                self.directory = directory
            while True:
                try:
                    os.makedirs(self.directory)
                    break
                except OSError:
                    self.directory += 'I'


        self.aggregation = defaultdict(lambda: defaultdict(OnlineVariance))
        self.round = 0

        self.plugin = plugin
        self.pluginargs = pluginargs


        self.panel_data = []
        self.write_panel_data_header = True

        if self.plugin is not None:
            self.plugin = self.plugin(*self.pluginargs)

    def put(self, msg):
        if msg[0] == 'snapshot_agg':
            _, round, group, data_to_write = msg
            if self.round == round:
                for key, value in data_to_write.items():
                    self.aggregation[group][key].update(value)
            else:
                self.make_aggregation_and_write()
                self.round = round
                for key, value in data_to_write.items():
                    self.aggregation[group][key].update(value)

        elif msg[0] == 'log':
            self.log(msg[1:])

        elif msg[0] == 'panel_log':
            (_,
              _str_round,
              group,
              _str_name,
              serial,
              data_to_write) = msg
            for action, data in data_to_write.items():
                self.log((_str_round, group, _str_name, '%s_%s' % (action, serial), data))

        elif msg[0] == 'trade_log':
            print("trade_log non implemented")

        else:
            try:
                getattr(self.plugin, msg[0])(*msg[1], **msg[2])
            except AttributeError:
                raise AttributeError(
                    "abcEconomics_db error '%s' command unknown" % msg)

    def log(self, msg):
        self.panel_data.append(msg)
        if len(self.panel_data) % 100000:
            if getsizeof(self.panel_data) > 1000000000:
                self.dump()

    def dump(self):
        if self.panel_data:
            df = pd.DataFrame(self.panel_data)
            df.rename(columns={0: 'round', 1: 'group', 2: 'name', 3: 'var', 4: 'value'}, inplace=True)

            df.pivot_table(values='value', index=['round', 'group', 'name'],
                                             columns=['var']).to_csv(self.directory + '/data.csv',
                                             mode='a', header=self.write_panel_data_header)
            self.write_panel_data_header = False
            self.panel_data = []

    def finalize(self, data):
        self.close()
        self._write_description_file(data)

    def close(self):
        self.make_aggregation_and_write()
        self.dump()

        try:
            self.plugin.close()
        except AttributeError:
            pass

    def make_aggregation_and_write(self):
        for group, table in self.aggregation.items():
            result = {'round': self.round}
            for key, data in table.items():
                result[key + '_ttl'] = data.sum()
                result[key + '_mean'] = data.mean()
                result[key + '_std'] = data.std()
                self.log((self.round, group, None, '%s_%s' % (key, '_ttl'), data.sum()))
                self.log((self.round, group, None, '%s_%s' % (key, '_mean'), data.sum()))
                self.log((self.round, group, None, '%s_%s' % (key, '_sum'), data.std()))
            self.aggregation[group].clear()

    def _write_description_file(self, data):
        if self.directory is not None:
            with open(os.path.abspath(self.directory + '/description.txt'), 'w') as description:
                description.write(json.dumps(
                    data,
                    indent=4,
                    skipkeys=True,
                    default=lambda x: 'not_serializeable'))


class MultiprocessingDatabase(Database, multiprocessing.Process):
    def __init__(self, directory, name, in_sok, trade_log, plugin=None, pluginargs=[]):
        super().__init__(directory, name, trade_log, plugin, pluginargs)
        self.in_sok = in_sok

    def run(self):
        if self.plugin is not None:
            self.plugin = self.plugin(*self.pluginargs)
        self.panel_data = []
        self.write_panel_data_header = True

        while True:
            try:
                msg = self.in_sok.get(timeout=120)
            except queue.Empty:
                print("simulation.finalize() must be specified at the end of simulation")
                msg = self.in_sok.get()
            if msg == "close":
                self.close()
                break
            self.put(msg)

    def finalize(self, data):
        self.in_sok.put('close')
        while self.is_alive():
            time.sleep(0.05)
        self._write_description_file(data)



import abcEconomics
from ball import Ball
from killer import Killer
from victim import Victim


def main(processes, rounds):
    simulation = abcEconomics.Simulation(processes=processes)

    print('build Killer')
    killer = simulation.build_agents(Killer, 'killer', 1)
    print('build Victim')
    victims = simulation.build_agents(Victim, 'victim', rounds)
    print('build Victim loudvictim')
    loudvictims = simulation.build_agents(Victim, 'loudvictim', rounds)
    print('build AddAgent')
    balls = simulation.build_agents(Ball, 'ball', 0)

    for time in range(rounds):
        simulation.advance_round(time)
        assert len(balls) == time, len(balls)
        deads = killer.kill_silent()
        victims.delete_agents(deads)
        deads = killer.kill_loud()
        loudvictims.delete_agents(deads)
        victims.am_I_dead()
        loudvictims.am_I_dead()
        balls.create_agents(Ball, 1)

    simulation.finalize()
    del simulation


if __name__ == '__main__':
    main(processes=1, rounds=30)
    print('Iteration with 1 core finished')
    main(processes=4, rounds=20)
    print('Iteration with multiple processes finished')



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# abcEconomics is open-source software. If you are using abcEconomics for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
# pylint: disable=W0212, C0111


""" The best way to start creating a simulation is by copying the start.py
file and other files from 'abcEconomics/template' in https://github.com/AB-CE/examples.

To see how to create a simulation, read :doc:`ipython_tutorial`.


This is a minimal template for a start.py::

    from agent import Agent
    from abcEconomics import *


    simulation = Simulation(name='abcEconomics')
    agents = simulation.build_agents(Agent, 'agent', 2)
    for time in range(100):
        simulation.advance_round(time)
        agents.one()
        agents.two()
        agents.three()
    simulation.graphs()

Note two things are important: there must be either a

:func:`~abcEconomics.simulation.graphs` or a :func:`~abcEconomics.simulation.finalize` at the end
otherwise the simulation blocks at the end.
Furthermore, every round needs to be announced using simulation.advance_round(time),
where time is any representation of time.

"""
import re
import time
import random
import queue
import multiprocessing as mp
from collections import OrderedDict
from .logger import  MultiprocessingDatabase, Database
from .agent import Agent  # noqa: F401
from .group import Group
from .notenoughgoods import NotEnoughGoods  # noqa: F401
from .agents import Firm, Household  # noqa: F401
from .gui import gui, graph  # noqa: F401
from .singleprocess import SingleProcess
from .multiprocess import MultiProcess


class Simulation(object):
    """ This is the class in which the simulation is run. Actions and agents have to
    be added. Databases and resource declarations can be added. Then run
    the simulation.

    Args:
        name:
            name of the simulation

        random_seed (optional):
            a random seed that controls the random number of the simulation

        trade_logging:
            Whether trades are logged,trade_logging can be
            'group' (fast) or 'individual' (slow) or 'off'

        processes (optional):
            The number of processes that runs in parallel. Each process hosts
            a share of the agents.
            By default, if this parameter is not specified, `processes` is all
            your logical processor cores times two, using hyper-threading when available.
            For easy debugging, set processes to one and the simulation is
            executed without parallelization.
            Sometimes it is advisable to decrease the number of processes to
            the number of logical or even physical processor cores on your
            computer.
            **For easy debugging set processes to 1, this way only one agent
            runs at a time and only one error message is displayed**

        check_unchecked_msgs:
            check every round that all messages have been received with get_massages or get_offers.

        path:
            path for database use None to omit directory creation.

        dbplugin, dbpluginargs:
            database plugin, see :ref:`Database Plugins`

        Example::

            simulation = Simulation(name='abcEconomics',
                                    trade_logging='individual',
                                    processes=None)


    Example for a simulation::

        num_firms = 5
        num_households = 2000

        w = Simulation(name='abcEconomics',
                       trade_logging='individual',
                       processes=None)

        w.panel('firm', command='after_sales_before_consumption')

        firms = w.build_agents(Firm, 'firm', num_firms)
        households = w.build_agents(Household, 'household', num_households)

        all = firms + households

        for time in range(100):
            self.time = time
            endowment.refresh_services('labor', derived_from='labor_endowment', units=5)
            households.recieve_connections()
            households.offer_capital()
            firms.buy_capital()
            firms.production()
            if time == 250:
                centralbank.intervention()
            households.buy_product()
            all.after_sales_before_consumption()
            households.consume()

        w.finalize()
        w.graphs()
    """

    def __init__(self, name='abcEconomics', random_seed=None, trade_logging='off', processes=1, dbplugin=None,
                 dbpluginargs=[], path='auto', multiprocessing_database=False):
        """
        """
        try:
            name = simulation_name  # noqa: F821
        except NameError:
            pass

        self.agents_created = False
        self.resource_endowment = []

        self.trade_logging_mode = trade_logging
        if self.trade_logging_mode not in ['individual', 'group', 'off']:
            Exception("trade_logging can be "
                      "'group' (fast) or 'individual' (slow) or 'off'"
                      ">" + self.trade_logging_mode + "< not accepted")

        self.processes = mp.cpu_count() * 2 if processes is None else processes

        if processes == 1:
            self.scheduler = SingleProcess()
        else:
            self.scheduler = MultiProcess(processes)

        if processes > 1 or multiprocessing_database:
            manager = mp.Manager()
            self.database_queue = manager.Queue()

            self._db = MultiprocessingDatabase(
                path,
                name,
                self.database_queue,
                trade_log=self.trade_logging_mode != 'off',
                plugin=dbplugin,
                pluginargs=dbpluginargs)
            self._db.start()
        else:
            self._db = self.database_queue = Database(path,
                name,
                trade_log=self.trade_logging_mode != 'off',
                plugin=dbplugin,
                pluginargs=dbpluginargs)

        self.path = self._db.directory

        if random_seed is None or random_seed == 0:
            random_seed = time.time()
        random.seed(random_seed)

        self.sim_parameters = OrderedDict(
            {'name': name, 'random_seed': random_seed})
        self.clock = time.time()
        self._time = None
        """ The current time set with simulation.advance_round(time)"""
        self._groups = {}
        """ A list of all agent names in the simulation """

    @property
    def time(self):
        """ Set and get time for simulation and all agents """
        return self._time

    @time.setter
    def time(self, time):
        """ Set and get time for simulation and all agents """
        self.advance_round(time)

    def advance_round(self, time):
        self._time = time
        print("\rRound" + str(time))
        str_time = re.sub('[^0-9a-zA-Z_\-]', '', str(time))
        self.scheduler.advance_round(time, str_time)

    def finalize(self):
        """ simulation.finalize() must be run after each simulation. It will
        write all data to disk

        Example::

            simulation = Simulation(...)
            ...
            for r in range(100):
                simulation.advance_round(r)
                agents.do_something()
                ...

            simulation.finalize()
        """
        print('')
        print("time only simulation %6.2f" %
              (time.time() - self.clock))

        self._db.finalize(self.sim_parameters)

        try:
            self.pool.close()
            self.pool.join()
        except AttributeError:
            pass

        print("time with data %6.2f" %
              (time.time() - self.clock))

    def build_agents(self, AgentClass, group_name,
                     number=None,
                     agent_parameters=None,
                     **parameters):
        """ This method creates agents.

        Args:

            AgentClass:
                is the name of the AgentClass that you imported

            group_name:
                the name of the group, as it will be used in the action list
                and transactions. Should generally be lowercase of the
                AgentClass.

            number:
                number of agents to be created.

            agent_parameters:
                a list of dictionaries, where each agent gets one dictionary.
                The number of agents is the length of the list

            any other parameters:
                are directly passed to the agent

        Example::

         firms = simulation.build_agents(Firm, 'firm',
             number=simulation_parameters['num_firms'])
         banks = simulation.build_agents(Bank, 'bank',
                                         agent_parameters=[{'name': 'UBS'},
                                         {'name': 'amex'},{'name': 'chase'}
                                         **simulation_parameters,
                                         loanable=True)

         centralbanks = simulation.build_agents(CentralBank, 'centralbank',
                                                number=1,
                                                rounds=num_rounds)
        """
        assert number is None or agent_parameters is None, \
            'either set number or agent_parameters in build_agents'
        assert number is not None or agent_parameters is not None, \
            'please set either the number or agent_parameters in build_agents'
        assert group_name.isidentifier()

        if agent_parameters is None:
            agent_parameters = {}
        if parameters is None:
            parameters = {}
        if number is not None:
            agent_parameters = [{} for _ in range(number)]

        self.sim_parameters[group_name] = parameters

        group = Group(self, self.scheduler, None,
                      agent_arguments={'group': group_name,
                                       'trade_logging': self.trade_logging_mode,
                                       'database': self.database_queue})
        group.create_agents(AgentClass, agent_parameters=agent_parameters, **parameters)
        self.agents_created = True
        self._groups[group_name] = group
        return group

    def create_agents(self, AgentClass, group_name, simulation_parameters=None, agent_parameters=None, number=1):

        raise Exception("create_agents is depreciated for Group.create_agents")

    def create_agent(self, AgentClass, group_name, simulation_parameters=None, agent_parameters=None):
        raise Exception("create_agent is depreciated for Group.create_agents")

    def delete_agent(self, *ang):
        raise Exception("delete_agent is depreciated for create_agents")

    def delete_agents(self, group, ids):
        """ This deletes a group of agents. The model has to make sure that other
        agents are notified of the death of agents in order to stop them from corresponding
        with this agent. Note that if you create new agents
        after deleting agents the ID's of the deleted agents are reused.

        Args:
            group:
                group of the agent

            ids:
                a list of ids of the agents to be deleted in that group
        """
        group = self._groups[group]
        group.delete_agents(ids)

    def graph(self):
        """ after the simulation is run, graphs() shows graphs of all data
        collected in the simulation. Shows the same output as the @gui
        decorator shows.


        Example::

            simulation = Simulation(...)
            for r in range(100):
                simulation.advance_round(r)
                agents.do_something()
                ...

            simulation.graphs()
        """
        self.finalize()
        graph(self.path, self.sim_parameters)



import abcEconomics
import platform


class MyAgent(abcEconomics.Agent):
    def init(self, name):
        self.name = name

    def say(self):
        return self.name


def main(processes, rounds):
    sim = abcEconomics.Simulation()
    sim.advance_round(0)

    myagents = sim.build_agents(MyAgent, 'myagent', agent_parameters=[{'name': 'me'},
                                                                      {'name': 'you'},
                                                                      {'name': 'him'},
                                                                      {'name': ('firm', 0)}])

    names = ['me', 'you', 'him', ('firm', 0)]
    for name in names:
        assert list(myagents.by_name(name).say()) == [name], (
            list(myagents.by_name(name).say()), [name])

    assert list(myagents.by_name(name).say()) == [name], (
        list(myagents.by_names(names).say()), names)
    sim.finalize()


if __name__ == '__main__':
    main(processes=1, rounds=30)
    if (platform.system() != 'Windows' and
            platform.python_implementation() != 'PyPy'):
        main(processes=4, rounds=30)



import abcEconomics
import time


class LoggerTest(abcEconomics.Agent):
    def init(self, rounds):
        self.last_round = rounds - 1
        self.create('money', 50)
        self.create('cookies', 3)

    def one(self):
        for good, quantity in self.possessions().items():
            self.log(good, quantity)
        self.log('round_log', self.time)
        pass

    def two(self):
        pass

    def three(self):
        pass

    def clean_up(self):
        pass

    def all_tests_completed(self):
        if self.time == self.last_round:
            time.sleep(0.5)
            print('Check database whether logging succeeded')



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# abcEconomics is open-source software. If you are using abcEconomics for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
""" This is the agent's facility to send and receive messages. Messages can
either be sent to an individual with :meth:`messenger.Messenger.message` or to a group with
:meth:`messenger.Messenger.message_to_group`. The receiving agent can either get all messages
with  :meth:`messenger.Messenger.get_messages_all` or messages with a specific topic with
:meth:`messenger.Messenger.get_messages`.
"""
from collections import defaultdict
from pprint import pprint
from random import shuffle


class Message(object):
    __slots__ = ['sender', 'receiver', 'topic', 'content']

    def __init__(self, sender, receiver, topic, content):
        self.sender = sender
        self.receiver = receiver
        self.topic = topic
        self.content = content

    def __getitem__(self, item):
        return self.content[item]

    def __repr__(self):
        return "<{sender: %s; receiver: %s; topic: %s; content: %s}>" % (
            str(self.sender), str(self.receiver), self.topic, str(self.content))


class Messenger:
    def __init__(self, id, agent_parameters, simulation_parameters):
        super(Messenger, self).__init__(id, agent_parameters, simulation_parameters)
        self._msgs = {}
        self.inbox = []
        self._out = []

    def send_envelope(self, receiver, topic, content):
        """ sends an envelope to the agent, the envelope contains the message (content),
        sender, receiver and topic. Agents receive it
        at the beginning of next round with :meth:`~abcEconomicsagent.Messenger.get_messages` or
        :meth:`~abcEconomicsagent.Messenger.get_messages_all`.

        The message that arrives has the following properties::
        message.sender
        message.receiver
        message.topic
        message.content

        Important the content, when received is in message.content

        Args::

            receiver:
                The name of the receiving agent a tuple (group, id).
                e.G. ('firm', 15)

            topic:
                string, with which this message can be received

            content:
                variable, tuple, dictionary or class, that is send.

        Example::

            ... household_01 ...
            self.send_envelope('firm', 01, 'quote_sell', {'good':'BRD', 'quantity': 5})

            ... firm_01 - one subround later ...

            requests = self.get_messages('quote_sell')
            for req in requests:
                self.make_offer(req.sender, req.content['good'], reg['quantity'], self.price[req.good])

        Example2::

         self.send_envelope('firm', 01, 'm', "hello my message")

        """
        msg = Message(sender=self.name,
                      receiver=receiver,
                      topic=topic,
                      content=content)
        self.send(receiver, topic, msg)

    def get_messages(self, topic='m'):
        """ self.get_messages() returns all new messages send with :meth:`~abcEconomicsagent.Messenger.send`
        and :meth:`~abcEconomicsagent.Messenger.send_envelope`. The order is randomized. self.get_messages(topic) returns all
        messages with a particular topic.

        A message is a string with the message. You can also retrieve the sender
        by `message.sender_group` and `message.sender_id` and view the topic with
        'message.topic'. (see example)

        If you are sending a float or an integer you need to access the message
        content with `message.content` instead of only `message`.

        ! if you want to recieve a **float** or an **int**, you must msg.content

        Returns a message object:
            msg.content:
                returns the message content string, int, float, ...
            msg:
                returns also the message content, but only as a string
            sender_group:
                returns the group name of the sender
            sender_id:
                returns the id of the sender
            topic:
                returns the topic

        Example::

         ... agent_01 ...
         self.send_envelope('firm_01', 'potential_buyers', 'hello message')

         ... firm_01 - one subround later ...
         potential_buyers = get_messages('potential_buyers')
         for msg in potential_buyers:
            print('message: ', msg)
            print('message: ', msg.content)
            print('group name: ', msg.sender_group)
            print('sender id: ', msg.sender_id)
            print('topic: ', msg.topic)

        """
        try:
            shuffle(self._msgs[topic])
        except KeyError:
            self._msgs[topic] = []
        return self._msgs.pop(topic)

    def get_messages_all(self):
        """ returns all messages irregardless of the topic, in a dictionary by topic

        A message is a string with the message. You can also retrieve the sender
        by `message.sender_group` and `message.sender_id` and view the topic with
        'message.topic'. (see example)

        If you are sending a float or an integer you need to access the message
        content with `message.content` instead of only `message`.
        """
        ret = {}
        for key, messages in self._msgs.items():
            shuffle(messages)
            ret[key] = messages
        self._msgs.clear()
        return ret

    def _do_message_clearing(self):
        """ agent receives all messages and objects that have been send in this
        subround and deletes the offers that where retracted, but not executed.

        '_d': delete received that the issuing agent retract
        'abcEconomics_receive_accept': clears a made offer that was accepted by the other agent
        'abcEconomics_receive_reject': deletes an offer that the other agent rejected
        'abcEconomics_receive_good': recive a 'free' good from another party
        """
        for typ, msg in self.inbox:
            if typ == 'abcEconomics_propose_buy':
                self._open_offers_buy[msg.good][msg.id] = msg
            elif typ == 'abcEconomics_propose_sell':
                self._open_offers_sell[msg.good][msg.id] = msg
            elif typ == 'abcEconomics_receive_accept':
                offer = self._receive_accept(msg)
                if self.trade_logging == 2:
                    self._log_receive_accept_group(offer)
                elif self.trade_logging == 1:
                    self._log_receive_accept_agent(offer)
            elif typ == 'abcEconomics_receive_reject':
                self._receive_reject(msg)
            elif typ == 'abcEconomics_receive_good':
                self._inventory.haves[msg[0]] += msg[1]
            elif typ == 'abcEconomics_receive_quote':
                self._quotes[msg.id] = msg
            elif typ == '!d':
                if msg[0] == 'r':
                    del self._contracts_pay[msg[1]][msg[2]]
                if msg[0] == 'd':
                    del self._contracts_deliver[msg[1]][msg[2]]
            elif typ == 'abcEconomics_forceexecute':
                getattr(self, msg[0])(*msg[1:])
            else:
                self._msgs.setdefault(typ, []).append(msg)
        self.inbox.clear()

    def _post_messages(self, agents):
        for name, envelope in self._out:
            try:
                agents[name].inbox.append(envelope)
            except KeyError:
                print(envelope)
                raise KeyError("Receiver %s does not exist" % str(name))
        self._out.clear()

    def _post_messages_multiprocessing(self, num_processes):
        out = self._out
        self._out = defaultdict(list)
        return out

    def send(self, receiver, topic, msg):
        """ sends a message to a receiver
        The agents receives it at the begin of each subround.

        Args:

        sends a message to agent. Agents receive it
        at the beginning of next round with :meth:`~abcEconomicsagent.Messenger.get_messages`(topic) or
        :meth:`~abcEconomicsagent.Messenger.get_messages_all`.

        Args:
            receiver:
                to whom to send the message
            topic:
                topic, under which to receive the message
            msg:
                Some datatype with a message


        Example::

            ... household_01 ...
            self.send(('firm', 01), 'quote_sell', {'good':'BRD', 'quantity': 5, 'sender': self.name})

            ... firm_01 - one subround later ...

            requests = self.get_messages('quote_sell')

            for req in requests:
                self.make_offer(req['sender'], req['good'], reg['quantity'], self.price[req.good])

        Example2::

         self.send(('firm', 01), 'm', "hello my message")

        """
        self._out.append((receiver, (topic, msg)))

    def _send_multiprocessing(self, receiver, typ, msg):
        """ Is used to overwrite _send in multiprocessing mode.
        Requires that self._out is overwritten with a defaultdict(list) """
        self._out[hash(receiver) % self._processes].append(
            (receiver, (typ, msg)))

    def check_for_lost_messages(self):
        """ Checks whether there are any messages, or trade requests that have not been
        processed """
        for offer in list(self.given_offers.values()):
            if offer.made < self.time:
                print("in agent %s this offers have not been retrieved:" %
                      self.name_without_colon)
                for offer in list(self.given_offers.values()):
                    if offer.made < self.time:
                        print(offer.__repr__())
                raise Exception('%s_%i: There are offers have been made before'
                                'last round and not been retrieved in this'
                                'round get_offer(.)' % (self.group, self.id))

        if sum([len(offers) for offers in list(self._open_offers_buy.values())]):
            pprint(dict(self._open_offers_buy))
            raise Exception('%s_%i: There are offers an agent send that have '
                            'not been retrieved in this round get_offer(.)' %
                            (self.group, self.id))

        if sum([len(offers) for offers in list(self._open_offers_sell.values())]):
            pprint(dict(self._open_offers_sell))
            raise Exception('%s_%i: There are offers an agent send that have '
                            'not been retrieved in this round get_offer(.)' %
                            (self.group, self.id))

        if sum([len(offers) for offers in list(self._msgs.values())]):
            pprint(dict(self._msgs))
            raise Exception('(%s, %i): There are messages an agent send that '
                            'have not been retrieved in this round '
                            'get_messages(.)' % (self.group, self.id))



""" abcEconomics can be started with a gui or provide visual data output """
from .webtext import abcEconomicsdescription
from .plot_with_plotly import ppl
from .gui_with_dash import dashgui


def gui(parameter_mask, names=None, header=None, story=None,
        title="Agent-Based Computational Economics",
        texts=None, pages=None, histograms=None,
        serve=False, runtime='browser-X', truncate_rounds=0,
        hostname='0.0.0.0', port=80, pypy=None):
    """ gui is a decorator that can be used to add a graphical user interface
    to your simulation.

    Args:

        parameter_mask:
            a dictionary with the parameter name as key and an example value
            as value. Instead of the example value you can also put a tuple:
            (min, default, max)

            parameters can be:
                - float:
                    {'exponent': (0.0, 0.5, 1.1)}

                - int:
                    {'num_firms': (0, 100, 100000)}

                - dict or list, which should be strings of a dict or a
                  list (see example):
                    {'list_to_edit': "['brd', 'mlk', 'add']"}

                - a list of options:
                    {'several_options': ['opt_1', 'opt_2', 'opt_3']}

                - a string:
                    {'name': '2x2'}

        names (optional):
            a dictionary with the parameter name as key and an alternative
            text to be displayed instead.

        title:
            a string with the name of the simulation.

        header:
            html string for a bar on the top

        story:
            a dictionary with text to be displayed alongeside the graphs.
            Key must be the graphs title, value can be text or html.

        pages:
            A dictinoary with title as key and links to external websites
            as values, which are displayed on the right hand side.

        truncate_rounds:
            Does not display the initial x rounds, in the result graphs

        runtime:
            webbrowser to start the simulation in, can be 'xui' or python's
            webbrowser module's webrowser string.

        histograms:
            specifies in which round histograms are generated. If it is
            not specified rounds from the menu is used. Alternatively you can
            create 'histogram' parameter in parameter_mask.

        serve:
            If you run this on your local machine serve must be False.
            If used as a web server must be True

        hostname:
            Hostname if serve is active, defaults to '0.0.0.0'

        port:
            Port if serve is active, defaults to 80

        pypy:
            Name of the pypy interpreter to run abcEconomics super fast. e.G. 'pypy' or
            'pypy3'. The mainfile needs to be run with cpython e.G.:
            :code:`python3 start.py`

    Example::

        parameter_mask = {'name': 'name',
                          'random_seed': None,
                          'rounds': 40,
                          'num_firms': (0, 100, 100000),
                          'num_households': (0, 100, 100000),
                          'exponent': (0.0, 0.5, 1.1),
                          'several_options': ['opt_1', 'opt_2', 'opt_3']
                          'list_to_edit': "['brd', 'mlk', 'add']",
                          'dictionary_to_edit': "{'v1': 1, 'v2': 2}"}

        names = {'num_firms': 'Number of Firms'}

        @gui(parameter_mask, names,
             title="Agent-Based Computational Economics",
             serve=False)
        def main(simulation_parameters):
            parameter_list = eval(simulation_parameters['list_to_edit'])
            simulation = Simulation()
            firms = simulation.build_agents(Firm,
                simulation_parameters['num_firms'])
            households = simulation.build_agents(Household,
                simulation_parameters['num_households'])

            for r in range(simulation_parameters['rounds']):
                simulation.advance_round(r)
                firms.work()
                households.buy()

        if __name__ == '__main__':
            main(simulation_parameters)
    """
    parameter_mask = ({} if parameter_mask is None else parameter_mask)
    names = ({} if names is None else names)
    story = ({} if story is None else story)
    texts = ([abcEconomicsdescription] if texts is None else texts)
    pages = ([] if pages is None else pages)

    def inner(simulation):
        dashgui(simulation, parameter_mask, names, title, header, truncate_rounds, texts, pages)
        return lambda _: None
    return inner


def graph(path, parameter_mask=None, names=None):
    """ After the simulation simulation.graphs displays all logged data,
    this can not be use in conjuncture with @gui.

    Args:

        parameter_mask (optional):
            simulation parameters to display
        names (optional):
            a dictionary with the parameter name as key and an alternative
            text to be displayed instead.
    """
    names = ({} if names is None else names)

    ppl(path)



import pandas as pd
import cufflinks  # extends pandas to work with plotly
import dash
import dash_core_components as dcc
import dash_html_components as html


def dateparse(x):
    return pd.datetime.strptime(x, '%Y-%m-%d')


def generate_graph_layout(path, datetime=False):
    infile = path + '/data.csv'
    df = pd.read_csv(infile)
    try:
        df['round'].apply(int)
    except ValueError:
        df['round'] = pd.to_datetime(df['round'])

    groups = set(df['group'])

    graphs = {}

    for group in groups:
        graphs[group] = []

        group_df = df[df['group'] == group]

        for col in group_df.columns:
            if col not in ['round', 'name', 'group']:
                table = group_df.pivot_table(columns='name', values=col, index='round')
                if len(table) > 0:
                    graphs[group].append(html.H3(children=group + ' - ' + col))
                    panel_graph = dcc.Graph(id=group + '_' + col,
                                            figure=table.figure(kind='scatter', asFigure=True))

                    if len(table.columns) == 1:
                        graphs[group].append(panel_graph)
                    elif len(table.columns) > 1:
                        sum_graph = dcc.Graph(id=group + '_' + col + '_mean',
                                              figure=table.sum(1).figure(kind='scatter', asFigure=True))
                        mean_graph = dcc.Graph(id=group + '_' + col + '_mean',
                                               figure=table.mean(1).figure(kind='scatter', asFigure=True))
                        tabs = dcc.Tabs(id=group + '_' + col + '_tab', children=
                                        [dcc.Tab(label='panel', children=panel_graph, value=1),
                                         dcc.Tab(label='sum', children=sum_graph, value=2),
                                         dcc.Tab(label='mean', children=mean_graph, value=3)],
                                         value=1)
                        graphs[group].append(tabs)



    layout = html.Div(children=[
        html.H1(children='abcEconomics')] +
        [dcc.Tabs(id='graphs', children=[dcc.Tab(label=group, children=graphs[group]) for group in groups])])

    return layout


def ppl(path):
    app = dash.Dash()

    app.layout = generate_graph_layout(path)
    app.run_server(debug=True, use_reloader=False)



import abcEconomics


class Agent(abcEconomics.Agent):
    def actionA(self):
        print(self.name, 'actionA')

    def actionB(self):
        print(self.name, 'actionB')


def main(processes, rounds):
    sim = abcEconomics.Simulation(processes=processes)

    aagents = sim.build_agents(Agent, 'aagent', number=5)
    bagents = sim.build_agents(Agent, 'bagent', number=5)

    actionA = aagents.actionA
    actionB = bagents.actionB

    for r in range(rounds):
        sim.advance_round(r)
        (aagents.actionA + aagents.actionB)()
        (aagents.actionA + bagents.actionB)()
        (aagents.actionB + bagents.actionB)()
        aagents.actionA()
        bagents.actionB()
        (actionA + actionB)()

    sim.finalize()


if __name__ == '__main__':
    main(1, 1)
    main(4, 1)




# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# abcEconomics is open-source software. If you are using abcEconomics for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
""" This file contains functions to compare floating point variables to 0. All
variables in this simulation as in every computer programm are floating
point variables. Floating point variables are not exact. Therefore var_a == var_b
has no meaning. Further a variable that is var_c = 9.999999999999966e-30 is for our
purpose equal to zero, but var_c == 0 would lead to False.
:meth:`is_zero`, :meth:`is_positive` and :meth:`is_negative` work
around this problem by defining float epsilon and determine whether the variable is
sufficiently close to zero or not.

This file also defines the :exc:`tools.NotEnoughGoods`
"""
from __future__ import division
epsilon = 1 / 1000000


def is_zero(x):
    """ checks whether a number is sufficiently close to zero. All variables
    in abcEconomics are floating point numbers. Due to the workings of floating point
    arithmetic. If x is 1.0*e^-100 so really close to 0, x == 0 will be false;
    is_zero will be true.
    """
    return -epsilon < x < epsilon


def is_positive(x):
    """ checks whether a number is positive and sufficiently different from
    zero. All variables in abcEconomics are floating point numbers. Due to the workings
    of floating point arithmetic. If x is 1.0*e^-100 so really close to 0,
    x > 0 will be true, eventhough it is very very small;
    is_zero will be true.
    """
    return - epsilon <= x


def is_negative(x):
    """ see is positive """
    return x <= epsilon



import abcEconomics
import random


class Give(abcEconomics.Agent):
    def init(self, rounds):
        self.last_round = rounds - 1
        if self.id == 1:
            self.tests = {'all': False, 'topic': False}
        else:
            self.tests = {}

    def one(self):
        if self.id == 0:
            self.create('cookies', random.uniform(0, 10000))
            self.cookies = self['cookies']
            quantity = random.uniform(0, self['cookies'])
            self.give(('give', 1), 'cookies', quantity)
            assert self['cookies'] == self.cookies - quantity
            self.send_envelope(('give', 1), topic='tpc', content=quantity)

    def two(self):
        if self.id == 1:
            rnd = random.randint(0, 1)
            if rnd == 0:
                msg = self.get_messages_all()
                msg = msg['tpc']
                self.tests['all'] = True
                assert len(msg) == 1, len(msg)
            elif rnd == 1:
                msg = self.get_messages('tpc')
                self.tests['topic'] = True
                assert len(msg) == 1, len(msg)
            msg = msg[0]
            assert msg.content == self['cookies']
            assert msg.sender == ('give', 0), msg.sender
            assert msg.topic == 'tpc'
            assert msg.receiver == ('give', 1)

    def three(self):
        pass

    def clean_up(self):
        self.destroy('cookies')

    def all_tests_completed(self):
        if self.time == self.last_round and self.id == 0:
            assert all(self.tests.values(
            )), 'not all tests have been run; abcEconomics workes correctly, restart the unittesting to do all tests %s' % self.tests
            print('Test abcEconomics.give:\t\t\t\t\tOK')
            print('Test abcEconomics.message:\t\t\t\tOK')
            print('Test abcEconomics.get_messages:\t\t\t\tOK')
            print('Test abcEconomics.get_messages_all:\t\t\tOK')



cdef class OnlineVariance:
    cdef long n
    cdef double _mean
    cdef double M2

    cpdef clear(self)
    cpdef update(self, double x)
    cpdef double std(self)
    cpdef double sum(self)
    cpdef double mean(self)
    cpdef clear(self)




import abcEconomics
from production_multifirm import ProductionMultifirm
from utility_household import UtilityHousehold


def main(processes, rounds=5):
    s = abcEconomics.Simulation(processes=processes, name='production_consumption_test')
    print('build ProductionMultifirm')
    productionmultifirm = s.build_agents(ProductionMultifirm, 'productionmultifirm', 1, rounds=rounds)
    print('build UtilityHousehold')
    utilityhousehold = s.build_agents(UtilityHousehold, 'utilityhousehold', 5, rounds=rounds)

    all_agents = utilityhousehold + productionmultifirm

    for r in range(rounds):
        s.advance_round(r)
        productionmultifirm.production()
        utilityhousehold.consumption()
        all_agents.all_tests_completed()
    s.finalize()


if __name__ == '__main__':
    main(processes=1, rounds=5)
    print('Iteration with 1 core finished')
    main(processes=4, rounds=5)
    print('Iteration with multiple processes finished')



""" :code:`python -m abcEconomics.show` shows the simulation results in ./result/*  """
from .gui import graph
from .gui.gui_with_dash import newest_subdirectory

def show():
    graph(newest_subdirectory('./result', ''))


if __name__ == '__main__':
    show()



""" Copyright 2012 Davoud Taghawi-Nejad

 Module Author: Davoud Taghawi-Nejad

 abcEconomics is open-source software. If you are using abcEconomics for your research you are
 requested the quote the use of this software.

 Licensed under the Apache License, Version 2.0 (the "License"); you may not
 use this file except in compliance with the License and quotation of the
 author. You may obtain a copy of the License at
       http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 License for the specific language governing permissions and limitations under
 the License.
"""
# pylint: disable=W0212, C0111
from collections import ChainMap
import re


class SingleProcess(object):
    """ This is a container for all agents. It exists only to allow for multiprocessing with MultiProcess.
    """

    def __init__(self):
        self.agents = {}

    def add_agents(self, Agent, simulation_parameters, agent_parameters, default_sim_params, maxid):
        """appends an agent to a group """
        if isinstance(agent_parameters, int):
            agent_parameters = ([] for _ in range(agent_parameters))

        names = {}
        for id, ap in enumerate(agent_parameters, maxid):
            agent = Agent(id, ap, {**default_sim_params, **simulation_parameters})
            agent.init(**ChainMap(simulation_parameters, ap))
            agent._str_name = re.sub('[^0-9a-zA-Z_]', '', str(agent.name))
            names[agent.name] = agent.name
            assert agent.name not in self.agents, ('Two agents with the same name %s' % str(agent.name))
            self.agents[agent.name] = agent
        return names

    def delete_agents(self, names):
        for name in names:
            self.agents.pop(name, None)

    def do(self, names, command, args, kwargs):
        self.rets = []
        for name in names:
            ret = self.agents[name]._execute(command, args, kwargs)
            self.rets.append(ret)

    def post_messages(self, names):
        for name in names:
            self.agents[name]._post_messages(self.agents)
        return self.rets

    def advance_round(self, time, str_time):
        for agent in self.agents.values():
            agent._advance_round(time, str_time)

    def group_names(self):
        return self.agents.keys()



cdef class OnlineVariance:
    def __cinit__(self):
        self.clear()

    cpdef clear(self):
        self.n = 0
        self.M2 = 0.0
        self._mean = 0.0

    cpdef update(self, double x):
        cdef double delta
        cdef double delta2
        cdef int i

        self.n += 1

        delta = x - self._mean
        self._mean += delta / self.n
        delta2 = x - self._mean
        self.M2 += delta * delta2

    cpdef double std(self):
        cdef int _
        if self.n < 2:
            return 0.0
        else:
            return (self.M2 / (self.n - 1)) ** 0.5

    cpdef double mean(self):
        cdef int i
        return self._mean

    cpdef double sum(self):
        cdef int i
        return self._mean * self.n



import platform
import start_core_engine
import start_logging_test
import start_logging_friendly_names_test
import start_delete_create
import start_production_consumption
import start_combinable_actions
import start_return
import start_custom_database
import start_agent_indexing
import start_agent_numerical_indexing
import start_individual_access
import start_service
import start_transform
import start_messaging
import start_messaging_with_envelope


def run_test(name, test):
    print(name + " test, 1 core")
    test.main(processes=1, rounds=10)
    print('Iteration of %s testing with 1 core finished' % name)

    if (platform.system() != 'Windows' and platform.python_implementation() != 'PyPy'):
        print("%s test, 4 cores" % name)
        test.main(processes=4, rounds=10)
        print('Iteration of %s testing with multiple processes finished' % name)
    else:
        print("PYPY and windows: functions not tested with multi-processes")


if __name__ == '__main__':
    run_test("Logging", start_logging_test)
    run_test("Friendly Names Logging", start_logging_friendly_names_test)
    run_test("Core engine", start_core_engine)
    run_test("Returning and parameters", start_return)
    run_test("Create/Delete", start_delete_create)
    run_test("Production and consumption", start_production_consumption)
    run_test("Combinable actions", start_combinable_actions)
    run_test("Custom Database", start_custom_database)
    run_test("individual agent indexing", start_agent_indexing)
    run_test("individual agent indexing numerical", start_agent_numerical_indexing)
    run_test("start_individual_access", start_individual_access)
    run_test("Test service", start_service)
    run_test("Test Transform method", start_transform)
    run_test("Messaging", start_messaging)
    run_test("Messaging with envelope", start_messaging_with_envelope)



import abcEconomics
from tools import is_zero
import random


class QuoteBuy(abcEconomics.Agent):
    def init(self, rounds, cut_of):
        self.last_round = rounds - 1
        self.cut_of = cut_of
        if self.id == 1:
            self.tests = {'accepted': False,
                          'not_answered': False, 'partial': False}

    def one(self):
        """ Acts only if he is agent 0: sends an buy quote to agent 1 quote
        """
        if self.id == 0:
            self.create('money', random.uniform(0, 10000))
            self.money = self['money']
            self.price = random.uniform(0.0001, 1)
            quantity = random.uniform(0, self.money / self.price)
            self.quote_buy('quotebuy', 1, 'cookies', quantity, self.price)

    def two(self):
        """ Acts only if he is agent 1: recieves quotes and accepts;
        rejects; partially accepts and leaves quotes unanswerd.
        """
        if self.id == 1:
            self.create('cookies', random.uniform(0, 10000))
            cookies = self['cookies']
            if random.randint(0, 1) == 0:
                quotes = self.get_quotes('cookies')
            else:
                quotes = self.get_quotes_all()
                quotes = quotes['cookies']
            assert quotes
            for quote in quotes:
                if random.randint(0, 1) == 0:
                    self.tests['not_answered'] = True
                    continue
                if self['cookies'] >= quote['quantity']:
                    self.accept_quote(quote)
                    self.final_money = quote['price'] * quote['quantity']
                    assert self['cookies'] == cookies - quote['quantity']
                    self.tests['accepted'] = True
                else:
                    self.accept_quote_partial(
                        quote, self['cookies'])
                    assert is_zero(self['cookies'])
                    self.final_money = cookies * quote['price']
                    self.tests['partial'] = True

    def three(self):
        """
        """
        if self.id == 0:
            offers = self.get_offers('cookies')
            for offer in offers:
                self.accept(offer)

    def clean_up(self):
        self.destroy_all('money')
        self.destroy_all('cookies')
        if self.id == 1:
            self.final_money = self['money']

    def all_tests_completed(self):
        if self.round == self.last_round and self.id == 1:
            assert all(self.tests.values(
            )), 'not all tests have been run; abcEconomics workes correctly, restart the unittesting to do all tests %s' % self.tests
            print('Test abcEconomics.quote_buy:\t\t\t\t\tOK')



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# abcEconomics is open-source software. If you are using abcEconomics for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
The Household class extends the agent by giving him utility functions and the ability to consume goods.
"""
from abcEconomics import NotEnoughGoods
import operator
from functools import reduce
from ..trade import get_epsilon
epsilon = get_epsilon()


class Household:
    def consume(self, utility_function, input_goods):
        """ consumes input_goods returns utility according to the agent's
        utility function.

        A utility_function, has to be set before see
        py:meth:`~abcEconomicsagent.Household.create_cobb_douglas_utility_function` or
        manually; see example.

        Args:

            utility_function:
                A function that takes goods as parameters and returns
                a utility or returns (utility, left_over_dict). Where left_over_dict is
                a dictionary of all goods that are not completely consumed

            input goods dictionary or list:
                dictionary containing the amount of input good used consumed or
                a list of all goods that get completely consumed.
        Raises:
            NotEnoughGoods: This is raised when the goods are insufficient.

        Returns:
            The utility as a number. To log it see example.

        Example::

            def utility_function(car, cookies, bike):
                utility = car ** 0.5 * cookies ** 0.2 * bike ** 0.3
                cookies = 0  # cookies are consumed, while the other goods are not consumed
                return utility, locals()


            def utility_function(cake, cookies, bonbons):  # all goods get completely consumed
                utility = cake ** 0.5 * cookies ** 0.2 * bonbons ** 0.3
                return utility

            self.consumption_set = {'car': 1, 'cookies': 2000, 'bike':  2}
            self.consume_everything = ['car', 'cookies', 'bike']
            try:
                utility = self.consume(utility_function, self.consumption_set)
            except NotEnoughGoods:
                utility = self.consume(utility_function, self.consume_everything)
            self.log('utility': {'u': utility})

        """
        if not isinstance(input_goods, dict):
            input_goods = {good: self[good] for good in input_goods}

        utility_and_result = utility_function(**input_goods)

        try:
            utility, result = utility_and_result
        except TypeError:
            result = {}
            utility = utility_and_result

        for good, quantity in input_goods.items():
            if self._inventory.haves[good] - quantity + result.get(good, 0) < -epsilon:
                raise NotEnoughGoods

        for good, quantity in input_goods.items():
            self._inventory.haves[good] -= quantity

        for good, quantity in result.items():
            self._inventory.haves[good] += quantity

        return utility

    def create_cobb_douglas_utility_function(self, exponents):
        """ creates a Cobb-Douglas utility function

        Utility_functions are than used as an argument in consume_with_utility,
        predict_utility and predict_utility_and_consumption.

        Args:
            {'input1': exponent1, 'input2': exponent2 ...}: dictionary
            containing good names 'input' and correstponding exponents
        Returns:
            A utility_function that can be used in consume_with_utility etc.

        Example:
        self._utility_function = self.create_cobb_douglas({'bread' : 10, 'milk' : 1})
        self.produce(self.plastic_utility_function, {'bread' : 20, 'milk' : 1})
        """
        def utility_function(**goods):
            utility = reduce(operator.mul,
                             [goods[name] ** exponent
                              for name, exponent in exponents.items()])
            return utility
        return utility_function



import os
from collections import defaultdict
import csv


def to_csv(directory, dataset):
    os.chdir(directory)
    tables = dataset.tables
    panels = defaultdict(list)
    aggs = defaultdict(list)
    for table_name in tables:
        typ = table_name.split('___')[0]
        group = table_name.split('___')[1]
        if typ == 'panel':
            panels[group].append(table_name)
        elif typ == 'aggregate':
            aggs[group].append(table_name)
        elif typ == 'trade':
            print([ds for ds in dataset[table_name].all()][0])
            save_to_csv('trade_', '_trade', dataset)

    for group, tables in aggs.items():
        join_table(tables, group, 'round', 'aggregate', dataset)

    for group, tables in panels.items():
        join_table(tables, group, 'name, round', 'panel', dataset)
        create_aggregated_table(group, dataset)
    dataset.commit()

    for group in aggs:
        save_to_csv('aggregate', group, dataset)

    for group in panels:
        save_to_csv('panel', group, dataset)
        save_to_csv('aggregated', group, dataset)
    os.chdir('../..')


def create_aggregated_table(group, dataset):
    columns = ', '.join('AVG(%s) %s_mean, SUM(%s) %s_ttl' % (c, c, c, c)
                        for c in get_columns(dataset, 'panel_%s' % group))
    try:
        dataset.query("CREATE TABLE aggregated_%s AS "
                      "SELECT round, %s FROM panel_%s GROUP BY round ORDER BY cast(round as float);"
                      % (group, columns, group))
    except Exception:
        print('round not castable as float; default to unordered group by')
        dataset.query("CREATE TABLE aggregated_%s AS "
                      "SELECT round, %s FROM panel_%s GROUP BY round;"
                      % (group, columns, group))
    dataset.update_table('aggregated_%s' % group)


def join_table(tables, group, indexes, type_, dataset):
    for i, table_name in enumerate(tables):
        if i == 0:
            dataset.query("CREATE TEMPORARY TABLE temp0 AS "
                          "SELECT * FROM %s;" % table_name)
        else:
            redundant_columns = (set(dataset[table_name].columns) &
                                 set(dataset['temp%i' % (i - 1)].columns))
            dataset.query("CREATE TEMPORARY TABLE temp%i AS "
                          "SELECT temp%i.* %s "
                          "FROM temp%i LEFT JOIN %s using(%s)"
                          % (i, i - 1,
                             get_str_columns(dataset, table_name,
                                             redundant_columns),
                             i - 1, table_name, indexes))
            dataset.query("DROP TABLE temp%i;" % (i - 1))
        dataset.query("DROP TABLE %s" % table_name)

    dataset.query("CREATE TABLE %s_%s AS "
                  "SELECT * FROM temp%i" % (type_, group, i))
    dataset.update_table('panel_%s' % group)
    dataset.query("DROP TABLE temp%i;" % i)


def save_to_csv(prefix, group, dataset):
    table = dataset['%s_%s' % (prefix, group)]
    with open('%s_%s.csv' % (prefix, group), 'w') as outfile:
        outdict = csv.DictWriter(outfile, fieldnames=table.columns)
        outdict.writeheader()
        for row in table:
            outdict.writerow(row)


def get_str_columns(dataset, table_name, redundant_columns):
    ret = ', '.join([' %s ' % c
                     for c in dataset[table_name].columns
                     if c not in list(redundant_columns)])
    if ret == '':
        return ''
    else:
        return ', %s' % ret


def get_columns(dataset, table_name):
    return [c for c in dataset[table_name].columns
            if c not in ('index', 'name', 'round')]




# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# abcEconomics is open-source software. If you are using abcEconomics for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
""" This file defines the :exc:`tools.NotEnoughGoods` """


class NotEnoughGoods(Exception):
    """ Methods raise this exception when the agent has less goods than needed

    These functions (self.produce, self.offer, self.sell, self.buy)
    should be encapsulated by a try except block::

     try:
        self.produce(...)
     except NotEnoughGoods:
        alternative_statements()

    """

    def __init__(self, _agent_name, good, amount_missing):
        self.good = good
        self.amount_missing = amount_missing
        self.name = _agent_name
        Exception.__init__(self)

    def __str__(self):
        return repr(str(self.name) + " " + str(self.amount_missing) + " of good '" + str(self.good) + "' missing")



import abcEconomics
import platform


class MyAgent(abcEconomics.Agent):
    def init(self):
        pass

    def say(self):
        return self.id


def main(processes, rounds):
    sim = abcEconomics.Simulation()
    sim.advance_round(0)

    myagents = sim.build_agents(MyAgent, 'myagent', number=5)

    for id in range(5):
        assert list(myagents[id].say()) == [id], (list(myagents[id].say()), [id])
    sim.finalize()


if __name__ == '__main__':
    main(processes=1, rounds=30)
    if (platform.system() != 'Windows' and
            platform.python_implementation() != 'PyPy'):
        main(processes=4, rounds=30)



from buy import Buy
from sell import Sell
from give import Give
from logger_test import LoggerTest
from abcEconomics import Simulation


def main(processes, rounds):
    s = Simulation(processes=processes, name='unittest')

    print('build Buy')
    buy = s.build_agents(Buy, 'buy', 1000, rounds=rounds)
    print('build Sell')
    sell = s.build_agents(Sell, 'sell', 1000, rounds=rounds)
    print('build Give')
    give = s.build_agents(Give, 'give', 2, rounds=rounds)
    print('build LoggerTest')
    loggertest = s.build_agents(
        LoggerTest, 'loggertest', 1, rounds=rounds)

    all = buy + sell + give + loggertest

    for r in range(rounds):
        s.advance_round(r)
        for _ in range(5):
            buy.one()
            buy.two()
            buy.three()
            buy.clean_up()
        buy.panel_log(variables=['price'])
        for _ in range(5):
            sell.one()
            sell.two()
            sell.three()
            sell.clean_up()
        for _ in range(5):
            give.one()
            give.two()
            give.three()
            give.clean_up()
        for _ in range(5):
            loggertest.one()
            loggertest.two()
            loggertest.three()
            loggertest.clean_up()

        all.all_tests_completed()
    s.finalize()


if __name__ == '__main__':
    main(processes=1, rounds=5)
    print('Iteration with 1 core finished')
    main(processes=4, rounds=5)
    print('Iteration with multiple processes finished')



from abcEconomics.trade import get_epsilon
from collections import defaultdict
from abcEconomics.notenoughgoods import NotEnoughGoods
from .expiringgood import ExpiringGood

epsilon = get_epsilon()


def isclose(a, b):
    rel_tol = 1e-9
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), 0.0)


class Inventory(object):
    def __init__(self, name):
        self.haves = defaultdict(int)
        self._reserved = defaultdict(int)
        self.name = name
        self._expiring_goods = []
        self._perishable = []

    def create(self, good, quantity):
        """ creates quantity of the good out of nothing

        Use with care. As long as you use it only for labor and
        natural resources your model is macro-economically complete.

        Args:
            'good': is the name of the good
            quantity: number
        """
        assert quantity >= 0.0
        self.haves[good] += quantity

    def create_timestructured(self, good, quantity):
        """ creates quantity of the time structured good out of nothing.
        For example::

            self.creat_timestructured('capital', [10,20,30])

        Creates capital. 10 units are 2 years old 20 units are 1 year old
        and 30 units are new.

        It can also be used with a quantity instead of an array. In this
        case the amount is equally split on the years.::

            self.creat_timestructured('capital', 60)

        In this case 20 units are 2 years old 20 units are 1 year old
        and 20 units are new.

        Args:
            'good':
                is the name of the good

            quantity:
                an arry or number
        """
        length = len(self.haves[good].time_structure)
        for i in range(length):
            qty = quantity[i] if type(quantity) == list else quantity / length
            self.haves[good].time_structure[i] += qty

    def destroy(self, good, quantity=None):
        """ destroys quantity of the good. If quantity is omitted destroys all

        Use with care.

        Args::

            'good':
                is the name of the good
            quantity (optional):
                number

        Raises::

            NotEnoughGoods: when goods are insufficient
        """
        if quantity is None:
            self.haves[good] = 0
        else:
            assert quantity >= 0.0
            available = self.haves[good]
            if available < quantity - epsilon:
                raise NotEnoughGoods(self.name, good, quantity - available)
            self.haves[good] -= quantity

    def reserve(self, good, quantity):
        self._reserved[good] += quantity
        if self._reserved[good] > self.haves[good]:
            if isclose(self._reserved[good], self.haves[good]):
                self._reserved[good] = self.haves[good]
            else:
                self._reserved[good] -= quantity
                raise NotEnoughGoods(self.name, good, quantity - (self.haves[good] - self._reserved[good]))

    def rewind(self, good, quantity):
        self._reserved[good] -= quantity

    def commit(self, good, committed_quantity, final_quantity):
        self._reserved[good] -= committed_quantity
        self.haves[good] -= final_quantity

    def transform(self, ingredient, unit, product, quantity=None):
        if quantity is None:
            quantity = self.haves[ingredient]
        self.destroy(ingredient, quantity)
        self.create(product, float(unit) * quantity)

    def possession(self, good):
        return self.not_reserved(good)

    def not_reserved(self, good):
        """ returns how much of good an agent possesses.

        Returns:
            A number.

        possession does not return a dictionary for self.log(...), you can use self.possessions([...])
        (plural) with self.log.

        Example::

            if self['money'] < 1:
                self.financial_crisis = True

            if not(is_positive(self['money']):
                self.bankruptcy = True

        """
        return float(self.haves[good] - self._reserved[good])

    def reserved(self, good):
        """ returns how much of a good an agent has currently reseed to sell or buy.

        Returns:
            A number.

        possession does not return a dictionary for self.log(...), you can use self.possessions([...])
        (plural) with self.log.

        Example::

            if self['money'] < 1:
                self.financial_crisis = True

            if not(is_positive(self['money']):
                self.bankruptcy = True

        """
        return self._reserved[good]

    def possessions(self):
        """ returns all possessions """
        return {good: float(self.haves[good] - self._reserved[good]) for good in self.haves}

    def calculate_netvalue(self, prices):
        return sum(quantity * prices[name]
                   for name, quantity in self.items())

    def calculate_assetvalue(self, prices):
        return sum(quantity * prices[name]
                   for name, quantity in self.items()
                   if prices[name] > 0)

    def calculate_liablityvalue(self, prices):
        return sum(quantity * prices[name]
                   for name, quantity in self.items()
                   if prices[name] < 0)

    def calculate_valued_assets(self, prices):
        ret = {name: quantity
               for name, quantity in self.items()
               if prices[name] >= 0}
        return ret

    def calculate_valued_liablities(self, prices):
        ret = {name: quantity
               for name, quantity in self.items()
               if prices[name] < 0}
        return ret

    def _advance_round(self):
        # expiring goods
        for good in self._expiring_goods:
            self.haves[good]._advance_round()

        # perishing goods
        for good in self._perishable:
            if good in self.haves:
                self.destroy(good)

    def __getitem__(self, good):
        return self.haves[good]

    def _declare_expiring(self, good, duration):
        self.haves[good] = ExpiringGood(duration)
        self._expiring_goods.append(good)



import abcEconomics


class Endowment(abcEconomics.Agent):
    def init(self, rounds):
        self.last_round = rounds - 1
        self.labor_endowment = 0

    def Iconsume(self):
        assert self['labor'] == 5 * self.time, (self.time, 5 * self.time, self['labor'])
        self.labor_endowment = self.time + 1

    def all_tests_completed(self):
        print('Test s.refresh_services:\t\t\tOK')


def main(processes, rounds):
    s = abcEconomics.Simulation(processes=processes, name='unittest')

    endowment = s.build_agents(Endowment, 'endowment', 2, rounds=rounds)

    for r in range(rounds):
        s.time = r
        endowment.refresh_services('labor', derived_from='labor_endowment', units=5)
        endowment.Iconsume()
    endowment.all_tests_completed()
    s.finalize()


if __name__ == '__main__':
    main(processes=1, rounds=5)
    print('Iteration with 1 core finished')
    main(processes=4, rounds=5)
    print('Iteration with multiple processes finished')



import abcEconomics
from abcEconomics.agents import Firm

from tools import is_zero


class ProductionMultifirm(abcEconomics.Agent, Firm):
    def init(self, rounds):
        self.last_round = rounds - 1

        def production_function(a, b):
            consumption_good = max(a ** 2, a ** 0.5 * b)
            a = 0
            b = b * 0.9
            return locals()

        self.pf = production_function
        self.cd = self.create_cobb_douglas('consumption_good', 5, {'a': 2, 'b': 1})
        self.leontief = self.create_leontief('consumption_good', {'a': 3, 'b': 1})
        self.car = self.create_leontief('car', {'wheels': 4, 'chassi': 1})
        self.ces = self.create_ces('consumption_good', gamma=0.5, shares={
                                   'a': 0.25, 'b': 0.25, 'c': 0.5})
        self.ces_flexible = self.create_ces(
            'consumption_good', gamma=0.5, multiplier=2)

        def many_goods_pf(a, b, c):
            soft_rubber = a ** 0.25 * b ** 0.5 * c ** 0.25
            hard_rubber = a ** 0.1 * b ** 0.2 * c ** 0.01
            waste = b / 2
            a = 0
            b = b * 0.9
            c = c
            return locals()

        self.many_goods_pf = many_goods_pf

    def production(self):
        self.create('a', 2)
        self.create('b', 2)
        self.produce(self.pf, {'a': 1, 'b': 2})

        assert self['a'] == 1, self['a']
        assert self['b'] == 1.8, self['b']
        assert is_zero(self['consumption_good'] - 1. ** 0.5 * 2), \
            self['consumption_good']
        self.destroy('a')
        self.destroy('b')
        self.destroy('consumption_good')

        result = self.pf(**{'a': 10, 'b': 10})
        assert result['consumption_good'] == max(10 ** 2, 10 ** 0.5 * 10)
        assert result['a'] == 0, result['a']
        assert result['b'] == 9, result['b']

        self.create('a', 2)
        self.create('b', 2)
        self.produce(self.cd, {'a': 1, 'b': 2})

        assert self['a'] == 1, self['a']
        assert self['b'] == 0, self['b']
        assert self['consumption_good'] == 5 * \
            1 ** 2 * 2 ** 1, self['consumption_good']
        self.destroy('a')
        self.destroy('b')
        self.destroy('consumption_good')

        self.create('a', 2)
        self.create('b', 2)
        self.produce(self.leontief, {'a': 1, 'b': 2})

        assert self['a'] == 1, self['a']
        assert self['b'] == 0, self['b']
        assert self['consumption_good'] == min(
            1 * 3, 2 * 1), self['consumption_good']
        self.destroy('a')
        self.destroy('consumption_good')

        self.create('a', 10)
        self.create('b', 10)
        self.create('c', 10)
        self.produce(self.many_goods_pf, {'a': 1, 'b': 2, 'c': 5})

        assert self['a'] == 9, self['a']
        assert self['b'] == 9.8, self['b']
        assert self['c'] == 10, self['c']
        assert self['soft_rubber'] == 1 ** 0.25 * \
            2 ** 0.5 * 5 ** 0.25, self['soft_rubber']
        assert self['hard_rubber'] == 1 ** 0.1 * \
            2 ** 0.2 * 5 ** 0.01, self['hard_rubber']
        assert self['waste'] == 2 / 2, self['waste']
        self.destroy('a')
        self.destroy('b')
        self.destroy('c')
        self.destroy('soft_rubber')
        self.destroy('hard_rubber')
        self.destroy('waste')

        self.create('a', 2)
        self.create('b', 2)
        self.create('c', 4)
        self.produce(self.ces, {'a': 1, 'b': 2, 'c': 4})

        assert self['a'] == 1, self['a']
        assert self['b'] == 0, self['b']
        assert self['c'] == 0, self['c']
        expected = (0.25 * 1 ** 0.5 + 0.25 * 2 **
                    0.5 + 0.5 * 4 ** 0.5) ** (1 / 0.5)
        assert self['consumption_good'] == expected, (
            self['consumption_good'], expected)
        self.destroy('a', 1)
        self.destroy('consumption_good', expected)

        self.create('a', 2)
        self.create('b', 2)
        self.create('c', 2)
        self.create('d', 2)
        self.create('e', 2)
        self.produce(self.ces_flexible, {
                     'a': 1, 'b': 2, 'c': 2, 'd': 2, 'e': 2})

        assert self['a'] == 1, self['a']
        assert self['b'] == 0, self['b']
        assert self['c'] == 0, self['c']
        assert self['d'] == 0, self['d']
        assert self['e'] == 0, self['e']
        expected = (2 * (0.2 * 1 ** 0.5 + 0.2 * 2 ** 0.5 + 0.2 *
                    2 ** 0.5 + 0.2 * 2 ** 0.5 + 0.2 * 2 ** 0.5) **
                    (1 / 0.5))
        assert is_zero(self['consumption_good'] - expected), (
            self['consumption_good'], expected)
        self.destroy('a', 1)
        self.destroy('consumption_good', expected)

    def all_tests_completed(self):
        if self.time == self.last_round and self.id == 0:
            print('Test produce:                            \tOK')
            print('Test create_production_function_one_good:\tOK')
            print('Test create_cobb_douglas:                \tOK')
            print('Test leontief:                           \tOK')
            print('Test create_production_function_many_goods\tOK')
            print('Test predict_produce_output               \tOK')
            print('Test predict_produce_input                \tOK')
            print('Test net_value                            \tOK')
            print('Test predict_net_value                    \tOK')



import platform
import csv
import abcEconomics
try:
    from math import isclose
except ImportError:
    def isclose(a, b):
        return a - 0.000001 < b < a + 0.000001


class Agent(abcEconomics.Agent):
    def init(self, name):
        self.name = name

    def go(self):
        self.create('money', 0.1)
        self.i = self.id
        self.r = self.time[0] + 10 * self.time[1]
        self.log('li', self.i)
        self.log('lr', self.r)


def compare(to_compare, path, message, processes):

    the_path = (path + '/' + 'data.csv')
    if platform.system() == 'Windows':  # windows compatibility
        the_path = the_path[the_path.find('/') + 1:]

    with open(to_compare, 'r') as generatedf:
        generated = {}
        for row in csv.DictReader(generatedf):
            try:
                generated[(row['round'], row['name'])] = row
            except KeyError:
                generated[row['round']] = row
    with open(the_path, 'r') as orginialf:
        orginial = {}
        for row in csv.DictReader(orginialf):
            try:
                orginial[(row['round'], row['name'])] = row
            except KeyError:
                orginial[row['round']] = row
    for row in generated:
        for key in generated[row]:
            if key != 'index':
                if generated[row][key] != orginial[row][key]:
                    assert isclose(float(generated[row][key]), float(orginial[row][key])), (
                        key, generated[row][key], orginial[row][key])
    for row in orginial:
        for key in orginial[row]:
            if key != 'index':
                if generated[row][key] != orginial[row][key]:
                    assert isclose(float(generated[row][key]), float(orginial[row][key])), (
                        key, generated[row][key], orginial[row][key])


def main(processes, rounds):
    simulation = abcEconomics.Simulation(name='logging_test_friendly_names', processes=processes)

    agents = simulation.build_agents(Agent, 'friendly_agent',
                                     agent_parameters=[{'name': 'A'}, {'name': 'Davoud'}, {'name': "fred"}, {'name': "F12"}])

    for rnd in range(10):
        for r in range(10):
            simulation.advance_round((r, rnd))
            agents.go()
            agents.agg_log(variables=['i', 'r'], goods=['money'])
            agents.panel_log(variables=['i', 'r'], goods=['money'])
    simulation.finalize()

    compare('friendly_data.csv',
            simulation.path, 'named logging test\t\t',
            processes)


if __name__ == '__main__':
    main(processes=1, rounds=None)
    if (platform.system() != 'Windows' and platform.python_implementation() != 'PyPy'):
        main(processes=4, rounds=None)



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# abcEconomics is open-source software. If you are using abcEconomics for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
The :py:class:`abcEconomics.Agent` class is the basic class for creating your agents.
It automatically handles the possession of goods of an agent. In order to
produce/transforme goods you also need to subclass the :py:class:`abcEconomics.Firm` or
to create a consumer the :py:class:`abcEconomics.Household`.

For detailed documentation on:

Trading, see :doc:`Trade`

Logging and data creation, see :doc:`Database`.

Messaging between agents, see :doc:`Messenger`.
"""
from .logger import Logger
from .trade import Trade
from .messenger import Messenger
from .goods import Goods


class Agent(Logger, Trade, Messenger, Goods):
    """ Every agent has to inherit this class. It connects the agent to the
    simulation and to other agent. The :class:`abcEconomics.Trade`,
    :class:`abcEconomics.Logger` and :class:`abcEconomics.Messenger` classes are included.
    An agent can also inheriting from :class:`abcEconomics.Firm`,
    :class:`abcEconomics.FirmMultiTechnologies` or :class:`abcEconomics.Household` classes.

    Every method can return parameters to the simulation.

    For example::

        class Household(abcEconomics.Agent, abcEconomics.Household):
            def init(self, simulation_parameters, agent_parameters):
                self.num_firms = simulation_parameters['num_firms']
                self.type = agent_parameters['type']
                ...

            def selling(self):
                for i in range(self.num_firms):
                    self.sell('firm', i, 'good', quantity=1, price=1)

            ...
            def return_quantity_of_good(self):
                return['good']


        ...

        simulation = Simulation()
        households = Simulation.build_agents(household, 'household',
                                             parameters={...},
                                             agent_parameters=[{'type': 'a'},
                                                               {'type': 'b'}])
        for r in range(10):
            simulation.advance_round(r)
            households.selling()
            print(households.return_quantity_of_good())



    """

    def __init__(self, id, agent_parameters, simulation_parameters):
        """ Do not overwrite __init__ instead use a method called init instead.
        init is called whenever the agent are build.
        """
        super(Agent, self).__init__(id, agent_parameters, simulation_parameters)
        """ self.id returns the agents id READ ONLY"""
        # unpacking simulation_parameters
        group = simulation_parameters['group']
        start_round = simulation_parameters.get('start_round', None)

        self.name = (group, id)
        self.id = id
        """ self.name returns the agents name, which is the group name and the
        id
        """
        self.group = group
        """ self.group returns the agents group or type READ ONLY! """
        # TODO should be group_address(group), but it would not work
        # when fired manual + ':' and manual group_address need to be removed

        self.time = start_round
        """ self.time, contains the time set with simulation.advance_round(time)
            you can set time to anything you want an integer or
            (12, 30, 21, 09, 1979) or 'monday' """

    def init(self):
        """ This method is called when the agents are build.
        It can be overwritten by the user, to initialize the agents.
        Parameters are the parameters given to
        :py:meth:`abcEconomics.Simulation.build_agents`.

        Example::

            class Student(abcEconomics.Agent):
                def init(self, rounds, age, lazy, school_size):
                    self.rounds = rounds
                    self.age = age
                    self.lazy = lazy
                    self.school_size = school_size

                def say(self):
                    print('I am', self.age ' years old and go to a school
                    that is ', self.school_size')


            def main():
                sim = Simulation()
                students = sim.build_agents(Student, 'student',
                                            agent_parameters=[{'age': 12, lazy: True},
                                                              {'age': 12, lazy: True},
                                                              {'age': 13, lazy: False},
                                                              {'age': 14, lazy: True}],
                                            rounds=50,
                                            school_size=990)

        """
        print("Warning: agent %s has no init function" % self.group)

    def _advance_round(self, time, str_time):
        super()._advance_round(time)
        self._str_round = str_time
        self.time = time

        if self.conditional_logging:
            if self.time in self.log_rounds:
                self.log_this_round = True
            else:
                self.log_this_round = False

    def _execute(self, command, args, kwargs):
        self._do_message_clearing()
        self._begin_subround()
        ret = getattr(self, command)(*args, **kwargs)
        self._end_subround()
        self._reject_polled_but_not_accepted_offers()
        return ret

    def _begin_subround(self):
        """ Overwrite this to make abcEconomics plugins, that need to do
        something at the beginning of every subround """
        pass

    def _end_subround(self):
        """ Overwrite this to make abcEconomics plugins, that need to do
        something at the beginning of every subround """
        pass



from .inventory import Inventory
from .notenoughgoods import NotEnoughGoods


class Goods:
    """ Each agent can access his goods. self['good_name'] shows the quantity of goods of a certain type an agent
    owns. Goods can be a string or any other python object.
    """

    def __init__(self, id, agent_parameters, simulation_parameters):
        # unpack simulation_parameters
        group = simulation_parameters['group']

        self._inventory = Inventory((group, id))
        self._resources = []

    def refresh_services(self, service, derived_from, units=1):
        self.destroy(service)
        self.create(service, getattr(self, derived_from) * units)

    def possession(self, good):
        """ returns how much of good an agent possesses.

        Returns:
            A number.

        possession does not return a dictionary for self.log(...), you can use
        self.possessions([...]) (plural) with self.log.

        Example::

            if self['money'] < 1:
                self.financial_crisis = True

            if not(is_positive(self['money']):
                self.bancrupcy = True

        """
        print("depreciated use self[good] or self.not_reserved[good]")
        return self._inventory[good]

    def possessions(self):
        """ returns all possessions """
        return self._inventory.possessions()

    def create(self, good, quantity):
        """ creates quantity of the good out of nothing

        Use create with care, as long as you use it only for labor and
        natural resources your model is macro-economically complete.

        Args:
            'good': is the name of the good
            quantity: number
        """
        self._inventory.create(good, quantity)

    def not_reserved(self, good):
        """ Returns the amount of goods that are not reserved for a trade

        Args:
            good
        """
        return self._inventory.not_reserved(good)

    def destroy(self, good, quantity=None):
        """ destroys quantity of the good. If quantity is omitted destroys all

        Args::

            'good':
                is the name of the good
            quantity (optional):
                number

        Raises::

            NotEnoughGoods: when goods are insufficient
        """
        self._inventory.destroy(good, quantity)

    def transform(self, inputs, outputs):
        """ Transforms a dictionary of goods into a new dictionary of goods.
        Raises NotEnoughGoods exception if not enough input goods are available

        Args:
            inputs: dictionary of goods and quantities, that are used in the transformation
            outputs: dictionary of goods and quantities, that are created by the transformation

        Example::

            self.transform(inputs={'gold': 1, 'copper': 4}, outputs={'redgold': 5})
        """
        for good, quantity in inputs.items():
            if self._inventory.haves[good] < quantity:
                raise NotEnoughGoods(self.name, good, quantity - self._inventory.haves[good])
        for good, quantity in inputs.items():
            self._inventory.destroy(good, quantity)
        for good, quantity in outputs.items():
            self._inventory.create(good, quantity)

    def __getitem__(self, good):
        return self._inventory.haves[good]



import abcEconomics
from abcEconomics.agents import Household


class Victim(abcEconomics.Agent, Household):
    def init(self):
        self.count = 1

    def am_I_dead(self):
        if self.id < self.time:
            raise Exception("should be dead %i" % self.id)




from __future__ import division
from __future__ import print_function
import abcEconomics
from abcEconomics.agents import Firm


class BuyExpiringCapital(abcEconomics.Agent, Firm):
    def init(self, rounds):
        self.last_round = rounds - 1

    def one(self):
        if self.id == 0:
            self.create('money', 10)
            self.buy('buyexpiringcapital', 1,
                     good='xcapital', quantity=10, price=1)
            assert self.free('xcapital') == 0
            assert self.free('money') == 0

    def two(self):
        if self.id == 1:
            self.create('xcapital', 10)
            assert self['xcapital'] == 10
            assert self['money'] == 0
            offer = self.get_offers('xcapital')[0]
            self['xcapital'] == 10
            self.accept(offer)
            assert self['xcapital'] == 0
            assert self['money'] == 10

    def three(self):
        if self.id == 0:
            assert self['xcapital'] == 10
            self.destroy('xcapital', 10)
        elif self.id == 1:
            assert self['money'] == 10
            self.destroy('money', 10)

    def clean_up(self):
        pass

    def all_tests_completed(self):
        if self.round == self.last_round and self.id == 0:
            print("BuyExpiringCapital \tOK")



from collections import deque
from abcEconomics.notenoughgoods import NotEnoughGoods
from abcEconomics.trade import get_epsilon
epsilon = get_epsilon()


class ExpiringGood(object):
    """ A good that expires after X rounds """

    def __init__(self, duration):
        self.duration = duration
        self.time_structure = deque([0 for _ in range(duration)])  # old to new

    def _advance_round(self):
        del self.time_structure[0]
        self.time_structure.append(0)

    def __add__(self, other):
        if isinstance(other, ExpiringGood):
            self.time_structure = [
                s + o for s, o in zip(self.time_structure, other.time_structure)]
        else:
            self.time_structure[-1] += other
        return self

    def __radd__(self, other):
        return other + sum(self.time_structure)

    def __sub__(self, other):
        if isinstance(other, ExpiringGood):
            other = float(other)
        if sum(self.time_structure) < - epsilon:
            raise NotEnoughGoods
        for i in range(len(self.time_structure)):
            if other >= self.time_structure[i]:
                other -= self.time_structure[i]
                self.time_structure[i] = 0
            else:
                self.time_structure[i] -= other
                break
        return self

    def __rsub__(self, other):
        return other - sum(self.time_structure)

    def __mul__(self, other):
        return sum(self.time_structure) * other

    def __floordiv__(self, other):
        return sum(self.time_structure) // other

    def __div__(self, other):
        return sum(self.time_structure) / other

    def __mod__(self, other):
        return sum(self.time_structure) % other

    def __pow__(self, other):
        return sum(self.time_structure) ** other

    def __cmp__(self, other):
        a = sum(self.time_structure)
        return (a > other) - (a < other)

    def __int__(self):
        return int(sum(self.time_structure))

    def __long__(self):
        return int(sum(self.time_structure))

    def __float__(self):
        return float(sum(self.time_structure))

    def __repr__(self):
        return str(sum(self.time_structure))

    def __get__(self, instance, owner=None):
        return sum(self.time_structure)

    def __abs__(self):
        return abs(self.__float__())



import abcEconomics
from tools import is_zero
import random
from abcEconomics import NotEnoughGoods


class Sell(abcEconomics.Agent):
    def init(self, rounds):
        self.last_round = rounds - 1
        self.tests = {'accepted': False, 'rejected': False,
                      'partial': False, 'full_partial': False}
        if self.id == 1:
            self.tests['not_answered'] = False

    def one(self):
        if self.id % 2 == 0:
            self.create('cookies', random.uniform(0, 10000))
            self.cookies = self['cookies']
            self.price = random.uniform(0.0001, 1)
            quantity = random.uniform(0, self.cookies)
            self.offer = self.sell(receiver=('sell', self.id + 1),
                                   good='cookies', quantity=quantity, price=self.price)
            assert self.not_reserved('cookies') == self.cookies - quantity

    def two(self):
        if self.id % 2 == 1:
            self.create('money', random.uniform(0, 10000))
            money = self['money']
            oo = self.get_offers('cookies')
            assert oo, oo
            for offer in oo:
                if random.randrange(0, 10) == 0:
                    self.tests['not_answered'] = True
                    continue
                elif random.randrange(0, 10) == 0:
                    self.reject(offer)
                    assert self['money'] == money
                    assert self['cookies'] == 0
                    self.tests['rejected'] = True
                    break  # tests the automatic clean-up of polled offers
                try:
                    if random.randrange(2) == 0:
                        self.accept(offer)
                        assert self['cookies'] == offer.quantity
                        assert self['money'] == money - offer.quantity * offer.price
                        self.tests['accepted'] = True
                    else:
                        self.accept(offer, offer.quantity)
                        assert self['cookies'] == offer.quantity
                        assert self['money'] == money - offer.quantity * offer.price
                        self.tests['full_partial'] = True

                except NotEnoughGoods:
                    self.accept(offer, self['money'] / offer.price)
                    assert self['money'] < 0.00000001, self['money']
                    test = (self['money'] - money) - \
                        self['cookies'] / offer.price
                    assert test < 0.00000001, test
                    self.tests['partial'] = True

    def three(self):
        if self.id % 2 == 0:
            offer = self.offer
            if offer.status == 'rejected':
                assert is_zero(self.cookies - self['cookies'])
                self.tests['rejected'] = True
            elif offer.status == 'accepted':
                if offer.final_quantity == offer.quantity:
                    assert self.cookies - \
                        offer.quantity == self['cookies']
                    assert self['money'] == offer.quantity * offer.price
                    self.tests['accepted'] = True
                else:
                    test = (self.cookies - offer.final_quantity) - \
                        self['cookies']
                    assert is_zero(test), test
                    test = self['money'] - \
                        offer.final_quantity * offer.price
                    assert is_zero(test), test
                    self.tests['partial'] = True
                    self.tests['full_partial'] = True
            else:
                SystemExit('Error in sell')

    def clean_up(self):
        self.destroy('cookies')
        self.destroy('money')

    def all_tests_completed(self):
        if self.time == self.last_round and self.id == 0:
            assert all(self.tests.values(
            )), 'not all tests have been run; abcEconomics workes correctly, restart the unittesting to do all tests %s' % self.tests
            print('Test abcEconomics.buy:\t\t\t\t\tOK')
            print('Test abcEconomics.accept\t(abcEconomics.buy):\t\tOK')
            print('Test abcEconomics.reject\t(abcEconomics.buy):\t\tOK')
            print('Test abcEconomics.accept\t(abcEconomics.buy):\tOK')
            print('Test reject pending automatic \t(abcEconomics.buy):\tOK')



import abcEconomics


class MyAgent(abcEconomics.Agent):
    def init(self):
        self.create('input_good', 10)

    def produce(self):
        assert self['input_good'] == 10 - self.time, (self.time, self['input_good'])
        assert self['output_good'] == 2 * self.time, (self.time, self['output_good'])
        self.transform({'input_good': 1}, {'output_good': 2})

    def all_tests_completed(self):
        print('Test goods.transform:\t\t\tOK')


def main(processes, rounds):
    sim = abcEconomics.Simulation(processes=processes)

    myagents = sim.build_agents(MyAgent, 'myagent', number=2)

    for time in range(rounds):
        sim.time = time
        myagents.produce()

    myagents.all_tests_completed()
    sim.finalize()


if __name__ == '__main__':
    main(processes=1, rounds=5)
    print('Iteration with 1 core finished')
    main(processes=4, rounds=5)
    print('Iteration with multiple processes finished')



# Copyright 2012 Davoud Taghawi-Nejad
#
#  Module Author: Davoud Taghawi-Nejad
#
#  abcEconomics is open-source software. If you are using abcEconomics for your research you are
#  requested the quote the use of this software.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may not
#  use this file except in compliance with the License and quotation of the
#  author. You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations under
# the License.


class Chain:
    def __init__(self, iterables):
        self.iterables = iterables

    def __iter__(self):
        for it in self.iterables:
            for element in it:
                yield element

    def __repr__(self):
        return repr(list(self.iterables))

    def __str__(self):
        return str(list(self.iterables))

    def __getitem__(self, item):
        try:
            return self.iterables[item]
        except IndexError:
            self.iterables = [i for i in iter(self)]
            return self.iterables[item]


class Action:
    # This allows actions of Group to be combined. For example::
    #
    #     (firms.sell & households.buy)()
    #
    # It works by returning a callable and combinable action from
    # the groups __getattr__ method.

    def __init__(self, scheduler, actions):
        self._scheduler = scheduler
        self.actions = actions

    def __add__(self, other):
        return Action(self._scheduler, self.actions + other.actions)

    def __call__(self, *args, **kwargs):
        for names, command, _, __ in self.actions:
            self._scheduler.do(names, command, args, kwargs)
        return Chain([self._scheduler.post_messages(action[0]) for action in self.actions])
        # itertools.chain, does not work here


class Group:
    """ A group of agents. Groups of agents inherit the actions of the agents class they are created by.
    When a group is called with an agent action all agents execute this actions simultaneously.
    e.G. :code:`banks.buy_stocks()`, then all banks buy stocks simultaneously.

    agents groups are created like this::

        sim = Simulation()

        Agents = sim.build_agents(AgentClass, 'group_name', number=100, param1=param1, param2=param2)
        Agents = sim.build_agents(AgentClass, 'group_name',
                                  param1=param1, param2=param2,
                                  agent_parameters=[dict(ap=ap1_agentA, ap=ap2_agentA),
                                                    dict(ap=ap1_agentB, ap=ap2_agentB),
                                                    dict(ap=ap1_agentC, ap=ap2_agentC)])

    Agent groups can be combined using the + sign::

        financial_institutions = banks + hedgefunds
        ...
        financial_institutions.buy_stocks()

    or::

       (banks + hedgefunds).buy_stocks()

    Simultaneous execution means that all agents act on the same information set and influence each other
    only after this action.

    individual agents in a group are addressable, you can also get subgroups (only from non combined groups):

        banks[5].buy_stocks()

        (banks[6,4] + hedgefunds[7,9]).buy_stocks()


    agents actions can also be combined::

        buying_stuff = banks.buy_stocks & hedgefunds.buy_feraries
        buy_stocks()

    or::


        (banks.buy_stocks & hedgefunds.buy_feraries)()

    """

    def __init__(self, sim, scheduler, names, agent_arguments=None):
        self.sim = sim
        self.num_managers = sim.processes
        self._scheduler = scheduler
        if names is None:
            self.names = set()
        else:
            self.names = names
        self._agent_arguments = agent_arguments
        self.panel_serial = 0
        self.last_action = "Begin_of_Simulation"
        self.num_agents = 0
        if agent_arguments is not None:
            self.agent_name_prefix = agent_arguments['group']

    def __add__(self, other):
        return Group(self.sim, self._scheduler, self.names.union(other.names))

    def __radd__(self, g):
        if isinstance(g, Group):
            return self.__add__(g)
        else:
            return self

    def panel_log(self, variables=[], goods=[], func={}, len=[]):
        """ panel_log(.) writes a panel of variables and goods
        of a group of agents into the database, so that it is displayed
        in the gui.

        Args:
            goods (list, optional):
                a list of all goods you want to track as 'strings'
            variables (list, optional):
                a list of all variables you want to track as 'strings'
            func (dict, optional):
                accepts lambda functions that execute functions. e.G.
                :code:`func = lambda self: self.old_money - self.new_money`
            len (list, optional):
                records the length of the list or dictionary with that name.

        Example in start.py::

            for round in simulation.next_round():
                firms.produce_and_sell()
                firms.panel_log(goods=['money', 'input'],
                            variables=['production_target', 'gross_revenue'])
                households.buying()
        """
        self._do('_panel_log', variables, goods, func, len, self.last_action)

    def agg_log(self, variables=[], goods=[], func={}, len=[]):
        """ agg_log(.) writes a aggregate data of variables and goods
        of a group of agents into the database, so that it is displayed
        in the gui.

        Args:
            goods (list, optional):
                a list of all goods you want to track as 'strings'
            variables (list, optional):
                a list of all variables you want to track as 'strings'
            func (dict, optional):
                accepts lambda functions that execute functions. e.G.
                :code:`func = lambda self: self.old_money - self.new_money`
            len (list, optional):
                records the length of the list or dictionary with that name.

        Example in start.py::

            for round in simulation.next_round():
                firms.produce_and_sell()
                firms.agg_log(goods=['money', 'input'],
                            variables=['production_target', 'gross_revenue'])
                households.buying()
        """
        self._do('_agg_log', variables, goods, func, len)

    def create_agents(self, Agent, number=1, agent_parameters=None, **common_parameters):
        """ Create new agents to this group. Works only for non-combined groups

        Args:
            Agent:
                The class used to initialize the agents
            agent_parameters:
                List of dictionaries of agent_parameters

            number:
                number of agents to create if agent_parameters is not set

            any keyword parameter:
                parameters directly passed to :code:`agent.init` methood

        Returns:
            The id of the new agent
        """
        if agent_parameters is None:
            agent_parameters = number

        new_names = self._scheduler.add_agents(Agent, common_parameters, agent_parameters, self._agent_arguments, self.num_agents)
        self.num_agents += len(new_names)
        self.names.update(new_names)
        return new_names

    def _do(self, command, *args, **kwargs):
        """ agent actions can be executed by :code:`group.action(args=args)`. """
        self.last_action = command
        return self._scheduler.do(self.names, command, args, kwargs)

    def __getattr__(self, command, *args, **kwargs):
        self.last_action = command
        return Action(self._scheduler, [(self.names, command, args, kwargs)])

    def delete_agents(self, names):
        """ Remove an agents from a group, by specifying their id.

        Args:
            ids:
                list of ids of the agent

        Example::

            students.delete_agents([1, 5, 15])
        """
        for name in names:
            self.names.remove(name)
        self._scheduler.delete_agents(names)

    def __getitem__(self, ids):
        try:
            names = {(self.agent_name_prefix, id) for id in ids}
        except TypeError:
            names = {(self.agent_name_prefix, ids)}
        return Group(self.sim, self._scheduler, names, self._agent_arguments)

    def by_names(self, names):
        """ Return a callable group of agents from a list of names.group

        Example::

            banks.by_names(['UBS', 'RBS', "DKB"]).give_loans() """
        names = set(names)
        return Group(self.sim, self._scheduler, names, self._agent_arguments)

    def by_name(self, name):
        """ Return a group of a single agents by its name """
        names = {name}
        return Group(self.sim, self._scheduler, names, self._agent_arguments)

    def __len__(self):
        """ Returns the length of a group """
        return len(self.names)

    def __repr__(self):
        return repr(self.names)
